

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
"""
Using the RU logging module can lead to deadlocks when used in multiprocess and
multithreaded environments.  This is due to the native python logging library
not being thread save.  See [1] for a Python bug report on the topic from 2009.
As of 2016, a patch is apparently submitted, but not yet accepted, for Python
3.5.  So we have to solve the problem on our own.

The fix is encoded in the 'after_fork()' method.  That method *must* be called
immediately after a process fork.  A fork will carry over all locks from the
parent process -- but it will not carry over any threads.  That means that, if
a fork happens while a thread owns the lock, the child process will start with
a locked lock -- but since there are no threads, there is actually nobody who
can unlock the lock, thus badaboom.

Since after the fork we *know* the logging locks should be unset (after all, we
are not in a logging call right now, are we?), we pre-emtively unlock them here.
But, to do so we need to know what locks exist in the first place.  For that
purpose, we create a process-singletone of all the loggers we hand out via
'get_logger()'.

Logging locks are 'threading.RLock' instances.  As such they can be locked
multiple times (from within the same thread), and we have to unlock them that
many times.  We use a shortcut, and create a new, unlocked lock.
"""

# ------------------------------------------------------------------------------
#
import os
import sys
import time
import types
import threading

from .atfork import *


# ------------------------------------------------------------------------------
#
class _LoggerRegistry(object):

    from .singleton import Singleton
    __metaclass__ = Singleton

    def __init__(self):
        self._registry = list()

    def add(self, logger):
        self._registry.append(logger)

    def release_all(self):
        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.lock = threading.RLock()
                  # handler.reset()
                logger = logger.parent

    def close_all(self):
        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.close()
                    logger.removeHandler(handler)
                logger = logger.parent
        self._registry = list()


# ------------------------------------------------------------------------------
#
_logger_registry = _LoggerRegistry()


# ------------------------------------------------------------------------------
#
def _after_fork():

    _logger_registry.release_all()
    logging._lock = threading.RLock()


# ------------------------------------------------------------------------------
#
def _atfork_prepare():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_parent():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_child():
    _after_fork()


# ------------------------------------------------------------------------------
#
atfork(_atfork_prepare, _atfork_parent, _atfork_child)


#
# ------------------------------------------------------------------------------


from   logging  import DEBUG, INFO, WARNING, WARN, ERROR, CRITICAL
import logging
import colorama

from .misc      import import_module
from .reporter  import Reporter


# The 'REPORT' level is used for demo output and the like.
# log.report.info(msg) style messages always go to stdout, and are enabled if:
#   - log level is set exactly to 'REPORT' (35)
#   - log level is < than 'REPORT' (ie. DEBUG, INFO, REPORT), AND 'stdout' is
#     not in the log target list
#
# To redirect debug logging to some file and have the reporter enabled, use:
#   - export RADICAL_PILOT_VERBOSE=DEBUG
#   - export RADICAL_PILOT_LOG_TGT=rp.log
#
# To interleave both log and reporter on screen, set the log target to 'stderr'.

_DEFAULT_LEVEL = 'ERROR'
REPORT = 35
OFF    = -1


# ------------------------------------------------------------------------------
#
def log_shutdown():
  # print 'shut my baby down'
    _logger_registry.close_all()
    logging.shutdown()


# ------------------------------------------------------------------------------
#
class ColorStreamHandler(logging.StreamHandler):
    """
    A colorized output SteamHandler
    """

    colours = {'DEBUG'    : colorama.Fore.CYAN,
               'INFO'     : colorama.Fore.GREEN,
               'WARN'     : colorama.Fore.YELLOW,
               'WARNING'  : colorama.Fore.YELLOW,
               'ERROR'    : colorama.Fore.RED,
               'CRITICAL' : colorama.Back.RED + colorama.Fore.WHITE
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, target):

        logging.StreamHandler.__init__(self, target)
        self._tty  = self.stream.isatty()
        self._term = getattr(self, 'terminator', '\n')


    # --------------------------------------------------------------------------
    #
    def emit(self, record):

        # only write in color when using a tty
        if self._tty:
            self.stream.write(self.colours[record.levelname]
                             + self.format(record)
                             + colorama.Style.RESET_ALL
                             + self._term)
        else:
            self.stream.write(self.format(record) + self._term)

      # self.flush()


# ------------------------------------------------------------------------------
#
class FSHandler(logging.FileHandler):

    def __init__(self, target):
        try:
            os.makedirs(os.path.abspath(os.path.dirname(target)))
        except:
            pass  # exists
        logging.FileHandler.__init__(self, target)


# ------------------------------------------------------------------------------
#
def get_logger(name, target=None, level=None, path=None, header=True):
    """
    Get a logging handle.

    'name'   is used to identify log entries on this handle.
    'target' is a comma separated list (or Python list) of specifiers, where
             specifiers are:
             '0'      : /dev/null
             'null'   : /dev/null
             '-'      : stdout
             '1'      : stdout
             'stdout' : stdout
             '='      : stderr
             '2'      : stderr
             'stderr' : stderr
             '.'      : logfile named ./<name>.log
             <string> : logfile named <string>
    'level'  log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    'header' print some version info on log open
    """

    if not name: name = 'radical'
    if not path: path = os.getcwd()

    if '/' in name:
        try:
            os.makedirs(os.path.normpath(os.path.dirname(name)))
        except:
            # dir exists
            pass

    logger = logging.getLogger(name)
    logger.propagate = False   # let messages not trickle upward
    logger.name      = name

    if logger.handlers:
        # we already conifgured that logger in the past -- just reuse it
        return logger

    # --------------------------------------------------------------------------
    # unconfigured loggers get configured.  We try to get the log level and
    # target from the environment.  We try env vars like this:
    #
    #     name  : radical.saga.pty
    #
    #     level : RADICAL_SAGA_PTY_VERBOSE
    #             RADICAL_SAGA_VERBOSE
    #             RADICAL_VERBOSE
    #
    #     target: RADICAL_SAGA_PTY_LOG_TARGET
    #             RADICAL_SAGA_LOG_TARGET
    #             RADICAL_LOG_TARGET
    # whatever is found first is applied.  Well, we actually try the way around
    # and then apply the last one -- but whatever ;)
    # Default level  is 'CRITICAL',
    # default target is '-' (stdout).


    env_name = name.upper().replace('.', '_')
    if '_' in env_name:
        elems = env_name.split('_')
    else:
        elems = [env_name]

    if not level:
        level = _DEFAULT_LEVEL
        for i in range(0,len(elems)):
            env_test = '_'.join(elems[:i + 1]) + '_VERBOSE'
            level    = os.environ.get(env_test, level)

    # backward compatible interpretation of SAGA_VERBOSE
    if env_name.startswith('RADICAL_SAGA'):
        level = os.environ.get('SAGA_VERBOSE', level)

    # translate numeric levels into symbolic ones
    level = {50 : 'CRITICAL',
             40 : 'ERROR',
             30 : 'WARNING',
             35 : 'REPORT',
             20 : 'INFO',
             10 : 'DEBUG',
              0 : _DEFAULT_LEVEL,
             -1 : 'OFF'}.get(level, str(level))

    # we want levels to be uppercase
    level = level.upper()

    level_warning = None
    if level not in ['OFF', 'DEBUG', 'INFO', 'WARN', 'WARNING', 'ERROR',
                     'CRITICAL', 'REPORT']:
        level_warning = "log level '%s' not supported -- reset to '%s'" \
                      % (level, _DEFAULT_LEVEL)
        level = _DEFAULT_LEVEL

    if level in ['REPORT']:
        level = REPORT

    if level in [OFF, 'OFF']:
        # we don't want logging actually
        target = 'null'

    if not target:
        target = 'stderr'
        for i in range(0,len(elems)):
            env_test = '_'.join(elems[:i + 1]) + '_LOG_TARGET'
            target   = os.environ.get(env_test, target)
            env_test = '_'.join(elems[:i + 1]) + '_LOG_TGT'
            target   = os.environ.get(env_test, target)

    if isinstance(target, list): targets = target
    else                       : targets = target.split(',')

    formatter = logging.Formatter('%(asctime)s: '
                                  '%(name)-20s: '
                                  '%(processName)-32s: '
                                  '%(threadName)-15s: '
                                  '%(levelname)-8s: '
                                  '%(message)s')

    # add a handler for each targets (using the same format)
    logger.targets = targets
    for t in logger.targets:
        if t in ['0', 'null']:
            handle = logging.NullHandler()
        elif t in ['-', '1', 'stdout']:
            handle = ColorStreamHandler(sys.stdout)
        elif t in ['=', '2', 'stderr']:
            handle = ColorStreamHandler(sys.stderr)
        elif t in ['.']:
            handle = FSHandler("%s/%s.log" % (path, name))
            t = ''  # dont add to name
        elif t.startswith('/'):
            handle = FSHandler(t)
            t = os.path.basename(t)
        else:
            handle = FSHandler("%s/%s" % (path, t))
        handle.setFormatter(formatter)
        if t:
            handle.name = '%s.%s' % (logger.name, str(t))
        else:
            handle.name = logger.name
        logger.addHandler(handle)

    if level in [OFF, 'OFF']:
        logger.setLevel('CRITICAL')
    else:
        logger.setLevel(level)

    if level_warning:
        logger.warn(level_warning)

    # if the given name points to a version or version_detail, log those
    if header:
        try:
            logger.info("%-20s version: %s", 'python.interpreter', ' '.join(sys.version.split()))
            tmp = import_module(name)
            if hasattr(tmp, 'version_detail'):
                logger.info("%-20s version: %s", name, getattr(tmp, 'version_detail'))
            elif hasattr(tmp, 'version'):
                logger.info("%-20s version: %s", name, getattr(tmp, 'version'))
        except:
            pass

        try:
            logger.info("%-20s pid: %s", '', os.getpid())
            logger.info("%-20s tid: %s", '', threading.current_thread().name)
        except:
            pass

    # --------------------------------------------------------------------------
    # add a close method to make sure we can close file handles etc.
    def _close(this_logger):
        while this_logger:
            for handler in this_logger.handlers:
                handler.close()
                this_logger.removeHandler(handler)
            this_logger = this_logger.parent
    # --------------------------------------------------------------------------
    logger.close = types.MethodType(_close,logger)


    # --------------------------------------------------------------------------
    # we also equip our logger with reporting capabilities, so that we can
    # report, for example, demo output whereever we have a logger.
    class _LogReporter(object):

        def __init__(self, logger):
            self._logger   = logger
            self._reporter = Reporter()

            # we always enable report if the log level is REPORT
            # otherwise, we enable report if log level  < REPORT and `targets`
            # does not contain `stdout`.
            if logger.getEffectiveLevel() == REPORT:
                self._enabled = True
            else:
                if  logger.getEffectiveLevel() < REPORT and \
                    '-'      not in logger.targets      and \
                    '1'      not in logger.targets      and \
                    'stdout' not in logger.targets      :
                    self._enabled = True
                else:
                    self._enabled = False

        def title(self, *args, **kwargs):
            if self._enabled:
                self._reporter.title(*args, **kwargs)

        def header(self, *args, **kwargs):
            if self._enabled:
                self._reporter.header(*args, **kwargs)

        def info(self, *args, **kwargs):
            if self._enabled:
                self._reporter.info(*args, **kwargs)

        def idle(self, *args, **kwargs):
            if self._enabled:
                self._reporter.idle(*args, **kwargs)

        def progress(self, *args, **kwargs):
            if self._enabled:
                self._reporter.progress(*args, **kwargs)

        def ok(self, *args, **kwargs):
            if self._enabled:
                self._reporter.ok(*args, **kwargs)

        def warn(self, *args, **kwargs):
            if self._enabled:
                self._reporter.warn(*args, **kwargs)

        def error(self, *args, **kwargs):
            if self._enabled:
                self._reporter.error(*args, **kwargs)

        def exit(self, *args, **kwargs):
            if self._enabled:
                self._reporter.exit(*args, **kwargs)

        def plain(self, *args, **kwargs):
            if self._enabled:
                self._reporter.plain(*args, **kwargs)

        def set_style(self, *args, **kwargs):
            self._reporter.set_style(args, kwargs)


    # we also equip our logger with reporting capabilities, so that we can
    # report, for example, demo output whereever we have a logger.
    logger.report = _LogReporter(logger)

    # keep the handle a round, for cleaning up on fork
    _logger_registry.add(logger)

    return logger


# -----------------------------------------------------------------------------
#
class LogReporter(object):
    """
    This class provides a wrapper around the Logger and (indirectly) Reporter
    classes, which adds uniform output filtering for the reporter while
    preserving the Reporter's API.
    """
    # --------------------------------------------------------------------------
    #
    def __init__(self, title=None, targets=['stdout'], name='radical',
                 level='REPORT'):

        self._logger = get_logger(name=name, target=targets, level=level)
        if title:
            self._logger.report.title(title)

        self.title     = self._logger.report.title
        self.header    = self._logger.report.header
        self.info      = self._logger.report.info
        self.progress  = self._logger.report.progress
        self.ok        = self._logger.report.ok
        self.warn      = self._logger.report.warn
        self.error     = self._logger.report.error
        self.exit      = self._logger.report.exit
        self.plain     = self._logger.report.plain
        self.set_style = self._logger.report.set_style


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    import radical.utils as ru
    r = ru.Reporter(title='test')

    r.header  ('header  \n')
    r.info    ('info    \n')
    r.progress('progress\n')
    r.ok      ('ok      \n')
    r.warn    ('warn    \n')
    r.error   ('error   \n')
    r.plain   ('plain   \n')

    r.set_style('error', color='yellow', style='ELTTMLE', segment='X')
    r.error('error ')

    i = 0
    j = 0
    for cname,col in r.COLORS.items():
        if cname == 'reset':
            continue
        i += 1
        for mname,mod in r.COLOR_MODS.items():
            if mname == 'reset':
                continue
            j += 1
            sys.stdout.write("%s%s[%12s-%12s] " % (col, mod, cname, mname))
            sys.stdout.write("%s%s" % (r.COLORS['reset'], r.COLOR_MODS['reset']))
        sys.stdout.write("\n")
        j = 0

    r.exit('exit\n', 2)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys
import time
import signal
import thread
import traceback

import Queue        as queue
import threading    as mt
import setproctitle as spt

from .logger  import get_logger


# ------------------------------------------------------------------------------
#
_ALIVE_MSG     = 'alive'  # message to use as alive signal
_START_TIMEOUT = 20.0     # time to wait for thread's startup signal.
                          # startup signal: 'alive' message
                          # is sent in both directions to ensure correct setup
_STOP_TIMEOUT  = 3.0      # time between temination signal and killing child
_BUFSIZE       = 1024*10  # default buffer size for socket recvs


# ------------------------------------------------------------------------------
#
class Thread(mt.Thread):
    '''
    This `Thread` class is a thin wrapper around `mt.Thread` which specifically
    manages the thread lifetime in a cautious and copperative way: *no* attempt
    on signal handling is made, no exceptions and interrupts are used, we expect
    to exclusively communicate between parent and child via a `mt.Events` and
    `Queue.Queue` instances.
   
    NOTE: At this point we do not implement the full `mt.Thread` constructor.
   
    The class also implements a number of initialization and finalization
    methods which can be overloaded by any deriving class.  While this can at
    first be a confusing richness of methods to overload, it significantly
    simplifies the implementation of non-trivial child threads.  By default,
    none of the initialized and finalizers needs to be overloaded.
   
    An important semantic difference are the `start()` and `stop()` methods:
    both accept an optional `timeout` parameter, and both guarantee that the
    child threads successfully started and completed upon return, respectively.
    If that does not happen within the given timeout, an exception is raised.
    Not that the child startup is *only* considered complete once all of its
    initialization methods have been completed -- the start timeout value must
    thus be chosen very carefully.  Note further that the *parent*
    initialization methods are also part of the `start()` invocation, and must
    also be considered when choosing the timeout value.  Parent initializers
    will only be executed once the child is known to be alive, and the parent
    can thus expect the child bootstrapping to be completed (avoiding the need
    for additional handshakes etc).  Any error in child or parent initialization
    will result in an exception, and the child will be terminated.
   
    Along the same lines, the parent and child finalizers are executed in the
    `stop()` method, prompting similar considerations for the `timeout` value.
   
    Any class which derives from this Thread class *must* overload the 'work_cb()`
    method.  That method is repeatedly called by the child thread's main loop,
    until:
      - an exception occurs (causing the child to fail with an error)
      - `False` is returned by `work_cb()` (causing the child to finish w/o error)
    '''

    # FIXME: assert that start() was called for some / most methods

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, target=None, log=None):

        # At this point, we only initialize members which we need before start
        self._ru_log = log  # ru.logger for debug output

        if not name:
            name = 'noname'

        # most importantly, we create an `mt.Event` and an `mt.Queue`.  Both
        # parent and child will watch the event, which thus acts as a lifeline
        # between the threads, to detect abnormal termination in threads.
        # The queue endpoint is used to send messages back and forth.
        self._ru_term     = mt.Event()
        self._ru_endpoint = queue.Queue()

        # NOTE: threadlocal storage *must* be defined on module level, and
        #       cannot be instance specific
        self._ru_local    = mt.local()   # keep some vars thread-local

        # note that some members are only really initialized when calling the
        # `start()`/`run()` methods, and the various initializers.
        self._ru_name              = name   # communicate name to thread
        self._ru_local.name        = name   # use for logfiles etc
        self._ru_local.started     = False  # start() has been called
        self._ru_local.spawned     = None   # set in start(), run()
        self._ru_local.initialized = False  # set to signal bootstrap success
        self._ru_local.terminating = False  # set to signal active termination
        self._ru_local.is_parent   = None   # set in start()
        self._ru_local.is_child    = None   # set in run()

        if not self._ru_log:
            # if no logger is passed down, log to null
            self._ru_log = get_logger('radical.util.threads')

        self._ru_log.debug('parent name: %s' % self._ru_local.name)

        if target:
            # we don't support `arguments`, yet
            self.work_cb = target

        # when cprofile is requested but not available, 
        # we complain, but continue unprofiled
        self._ru_cprofile = False
        if self._ru_name in os.environ.get('RADICAL_CPROFILE', '').split(':'):
            try:
                self._ru_log.error('enable cprofile for %s', self._ru_local.name)
                import cprofile
                self._ru_cprofile = True
            except:
                self._ru_log.error('cannot import cprofile - disable')

        # base class initialization
        super(Thread, self).__init__(name=self._ru_local.name)

        # we don't want threads to linger around, waiting for children to
        # terminate, and thus create sub-threads as daemons.
        self.daemon = True


    # --------------------------------------------------------------------------
    #
    def _ru_msg_send(self, msg):
        '''
        send new message to self._ru_endpoint.  We make sure that the
        message is not larger than the _BUFSIZE define in the RU source code.
        '''

        if not self._ru_local.spawned:
            # no child, no communication channel
            raise RuntimeError("can't communicate w/o child")

        self._ru_log.info('put message: [%s] %s', self._ru_local.name, msg)
        self._ru_endpoint.put('%s' % msg)


    # --------------------------------------------------------------------------
    #
    def _ru_msg_recv(self, timeout=None):
        '''
        receive a message from self._ru_endpoint.
        
        This call is non-blocking: if no message is available, return an empty 
        string.
        '''

        if not self._ru_local.spawned:
            # no child, no communication channel
            raise RuntimeError("can't communicate w/o child")

        try:
            if timeout:
                msg = self._ru_endpoint.get(block=True, timeout=timeout)
            else:
                msg = self._ru_endpoint.get(block=False)
            self._ru_log.info('get msg: %s', msg)
            return msg

        except queue.Empty:
            self._ru_log.warn('get msg timed out')
            return ''


    # --------------------------------------------------------------------------
    #
    def is_alive(self, strict=True):
        '''
        Since our threads are daemon threads we don't need to wait for them to
        actually die, but consider it sufficient to have the terminationm signal
        set.  However, that can potentially cause unexpected results when locks
        are used, as one would not expect to have locks hold by threads which
        are not alive.  For that reason we default to the 'true' (strict) alive
        notion of the native Python thread - but the non-strict version can be
        requates via `strict=False`.
        '''

        alive  = super(Thread, self).is_alive() 
        termed = self._ru_term.is_set()

        if strict:
            return alive
        else:
            return alive and not termed


    # --------------------------------------------------------------------------
    #
    def is_valid(self, term=True):
        '''
        This method provides a self-check, and will call `stop()` if that check
        fails and `term` is set.  If `term` is not set, the return value of
        `is_alive()` is passed through.
        
        The check will basically assert that `is_alive()` is `True`.   The
        purpose is to call this check frequently, e.g. at the begin of each
        method invocation and during longer loops, as to catch failing
        sub-thread and to then terminate.

        This method will always return True if `start()` was not called, yet.
        '''

        if not self._ru_local.initialized:
            return True

        alive = self.is_alive(strict=False)

        if not alive and term:
            self._ru_log.warn('alive check failed - stop [%s - %s]', alive, term)
            self.stop()
        else:
            return alive


    # --------------------------------------------------------------------------
    #
    def start(self, spawn=True, timeout=None):
        '''
        Overload the `mt.Thread.start()` method, and block (with timeout) until
        the child signals to be alive via a message over our queue.  Also
        run all initializers.

        If spawn is set to `False`, then no child thread is actually created,
        but the parent initializers will be executed nonetheless.

        '''

        self._ru_log.debug('start thread')

        if timeout is None:
            timeout = _START_TIMEOUT

        if spawn:
            # start `self.run()` in the child thread, and wait for it's
            # initalization to finish, which will send the 'alive' message.
            super(Thread, self).start()
            self._ru_local.spawned = True

        # this is the parent now - set role flags.
        self._ru_local.is_parent = True
        self._ru_local.is_child  = False

        # from now on we need to invoke `self.stop()` for clean termination.
        try: 

            if self._ru_local.spawned:
                # we expect an alive message message from the child, within
                # timeout
                #
                # NOTE: If the child does not bootstrap fast enough, the timeout
                #       will kick in, and the child will be considered dead,
                #       failed and/or hung, and will be terminated!  Timeout can
                #       be set as parameter to the `start()` method.
                msg = self._ru_msg_recv(timeout=timeout)

                if msg != _ALIVE_MSG:
                    raise RuntimeError('Unexpected child message from %s (%s)' \
                                      % (self._ru_local.name, msg))


            # When we got the alive messages, only then will we call the parent
            # initializers.  This way those initializers can make some
            # assumptions about successful child startup.
            self._ru_initialize()

        except Exception as e:
            self._ru_log.exception('%s initialization failed', self._ru_local.name)
            self.stop()
            raise

        # if we got this far, then all is well, we are done.
        self._ru_log.debug('child thread %s started', self._ru_local.name)

        # child is alive and initialized, parent is initialized - Wohoo!


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        This method MUST NOT be overloaded!

        This is the workload of the child thread.  It will first call the child
        initializers and then repeatedly call `self.work_cb()`, until being
        terminated.  When terminated, it will call the child finalizers and
        exit.  Note that the child will also terminate once `work_cb()` returns
        `False`.

        The implementation of `work_cb()` needs to make sure that this thread is
        not spinning idly -- if there is nothing to do in `work_cb()` at any point
        in time, the routine should at least sleep for a fraction of a second or
        something.

        The child thread will automatically terminate (incl. finalizer calls)
        when the parent thread dies. It is thus not possible to create orphaned
        threads -- which is an explicit purpose of this implementation.  
        '''

        # set local data
        self._ru_local.name = self._ru_name + '.thread'
        self._ru_log.debug('child name: %s' % self._ru_local.name)

        self._ru_local.started     = True   # start() has been called
        self._ru_local.spawned     = True   # set in start(), run()
        self._ru_local.initialized = False  # set to signal bootstrap success
        self._ru_local.terminating = False  # set to signal active termination
        self._ru_local.is_parent   = False  # set in start()
        self._ru_local.is_child    = True   # set in run()

        # if no profiling is wanted, we just run the workload and exit
        if not self._ru_cprofile:
            self._run()

        # otherwise we run under the profiler, obviously
        else:
            import cprofile
            cprofiler = cProfile.Profile()
            cprofiler.runcall(self._run)
            cprofiler.dump_stats('%s.cprof' % (self._ru_local.name))


    # --------------------------------------------------------------------------
    #
    def _run(self):

        # FIXME: ensure that this is not overloaded
        # TODO:  how?

        try:
            # we consider the invocation of the child initializers to be part of
            # the bootstrap process, which includes starting the watcher thread
            # to watch the parent's health (via the socket healt).
            try:
                self._ru_initialize()

            except BaseException as e:
                self._ru_log.exception('abort')
                self._ru_msg_send(repr(e))
                sys.stderr.write('initialization error in %s: %s\n'
                        % (self._ru_local.name, repr(e)))
                sys.stderr.flush()
                self._ru_term.set()

            # initialization done - we only now send the alive signal, so the
            # parent can make some assumptions about the child's state
            if not self._ru_term.is_set():
                self._ru_log.info('send alive')
                self._ru_msg_send(_ALIVE_MSG)

            # enter the main loop and repeatedly call 'work_cb()'.  
            #
            # If `work_cb()` ever returns `False`, we break out of the loop to call the
            # finalizers and terminate.
            #
            # In each iteration, we also check if the Event is set -- if this is
            # the case, we assume the parent to be dead and terminate (break the
            # loop).
            while not self._ru_term.is_set():
            
                # des Pudel's Kern
                if not self.work_cb():
                    self._ru_msg_send('work finished')
                    break

        except BaseException as e:

            # This is a very global except, also catches 
            # sys.exit(), keyboard interrupts, etc.  
            # Ignore pylint and PEP-8, we want it this way!
            self._ru_log.exception('abort: %s', repr(e))
            self._ru_msg_send(repr(e))

        try:
            # note that we always try to call the finalizers, even if an
            # exception got raised during initialization or in the work loop
            # initializers failed for some reason or the other...
            self._ru_finalize()

        except BaseException as e:
            self._ru_log.exception('finalization error')
            self._ru_msg_send('finalize(): %s' % repr(e))

        self._ru_msg_send('terminating')

        # all is done and said - begone!
        return


    # --------------------------------------------------------------------------
    #
    def stop(self, timeout=None):
        '''
        `stop()` is symetric to `start()`, in that it can only be called by the
        parent thread


        NOTE: `stop()` implies `join()`!  Use `terminate()` if that is not
              wanted.
        '''

        # FIXME: This method should reduce to 
        #           self.terminate(timeout)
        #           self.join(timeout)
        #        ie., we should move some parts to `terminate()`.

        if not hasattr(self._ru_local, 'name'):

            # This thread is now being stopped from *another* thread, ie.
            # neither the parent nor the child thread, so we can't access either
            # thread local storage.  we thus only set the termination signal
            # (which is accessible), and leave all other handling to somebody
            # else.
            #
            # Specifically, we will not be able to join this thread, and no
            # timeout is enforced
            self._ru_log.info('signal stop for %s - do not join', get_thread_name())
            self._ru_term.set()
            return

        if not timeout:
            timeout = _STOP_TIMEOUT

        # if stop() is called from the child, we just set term and leave all
        # other handling to the parent.
        if self._ru_local.is_child:
            self._ru_log.info('child calls stop()')
            self._ru_term.set()
          # cancel_main_thread()
            return

        self._ru_log.info('parent stops child')

        # make sure we don't recurse
        if self._ru_local.terminating:
            return
        self._ru_local.terminating = True

        # call finalizers - this sets `self._ru_term`
        self._ru_finalize()

        # After setting the termination event, the child should begin
        # termination immediately.  Well, whenever it realizes the event is set,
        # really.  We wait for that termination to complete.
        self.join()


    # --------------------------------------------------------------------------
    #
    def join(self, timeout=None):

      # raise RuntimeError('call stop instead!')
      #
      # we can't really raise the exception above, as that would break symmetry
      # with the Process class -- see documentation there.
      #
      # FIXME: not that `join()` w/o `stop()` will not call the parent 
      #        finalizers.  We should call those in both cases, but only once.

        if not hasattr(self._ru_local, 'name'):

            # This thread is now being stopped from *another* thread, ie.
            # neither the parent nor the child thread, so we can't access either
            # thread local storage.  we thus only set the termination signal
            # (which is accessible), and leave all other handling to somebody
            # else.
            #
            # Specifically, we will not be able to join this thread, and no
            # timeout is enforced
            self._ru_log.info('signal stop for %s - do not join', get_thread_name())
            self._ru_term.set()
            return

        if not timeout:
            timeout = _STOP_TIMEOUT

        if self._ru_local.is_parent:
            try:
                super(Thread, self).join(timeout=timeout)
            except Exception as e:
                self._ru_log.warn('ignoring %s' % e)


    # --------------------------------------------------------------------------
    #
    def _ru_initialize(self):
        '''
        Perform basic settings, then call common and parent/child initializers.
        '''

        try:
            # call parent and child initializers, respectively
            if self._ru_local.is_parent:
                self._ru_initialize_common()
                self._ru_initialize_parent()

                self.ru_initialize_common()
                self.ru_initialize_parent()

            elif self._ru_local.is_child:
                self._ru_initialize_common()
                self._ru_initialize_child()

                self.ru_initialize_common()
                self.ru_initialize_child()

            self._ru_local.initialized = True

        except Exception as e:
            self._ru_log.exception('initialization error')
            raise RuntimeError('initialize: %s' % repr(e))


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_common(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_parent(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_child(self):

        # TODO: should we also get an alive from parent?
        #
        # FIXME: move to _ru_initialize_common
        #

        self._ru_log.info('child (me) initializing')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_common(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, for both the parent and the child thread (individually).  If
        this fails on either side, the thread startup is considered failed.
        '''

        self._ru_log.debug('ru_initialize_common (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_parent(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, in the parent thread.  If this fails, the thread startup is
        considered failed.
        '''

        self._ru_log.debug('ru_initialize_parent (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_child(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, in the child thread.  If this fails, the thread startup is
        considered failed.
        '''
    
        self._ru_log.debug('ru_initialize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def _ru_finalize(self):
        '''
        Call common and parent/child initializers.  
        
        Note that finalizers are called in inverse order of initializers.
        '''

        try:
            # signal termination
            self._ru_term.set()

            # call parent and child finalizers, respectively
            if self._ru_local.is_parent:
                self.ru_finalize_parent()
                self.ru_finalize_common()

                self._ru_finalize_parent()
                self._ru_finalize_common()

            elif self._ru_local.is_child:
                self.ru_finalize_child()
                self.ru_finalize_common()

                self._ru_finalize_child()
                self._ru_finalize_common()

        except Exception as e:
            self._ru_log.exception('finalization error')
            raise RuntimeError('finalize: %s' % repr(e))


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_common(self):
    
        pass


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_parent(self):
    
        pass


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_child(self):
    
        pass


    # --------------------------------------------------------------------------
    #
    def ru_finalize_common(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or thread child termination, in the parent thread, in both
        the parent and the child thread (individually).
        '''

        self._ru_log.debug('ru_finalize_common (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_finalize_parent(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or thread child termination, in the parent thread.
        '''

        self._ru_log.debug('ru_finalize_parent (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_finalize_child(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or thread child termination, in the child thread.
        '''

        self._ru_log.debug('ru_finalize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def work_cb(self):
        '''
        This method MUST be overloaded.  It represents the workload of the
        thread, and will be called over and over again.

        This has several implications:

          * `work_cb()` needs to enforce any call rate limits on its own!
          * in order to terminate the child, `work_cb()` needs to either raise an
            exception, or call `sys.exit()` (which actually also raises an
            exception).

        Before the first invocation, `self.ru_initialize_child()` will be called.
        After the last invocation, `self.ru_finalize_child()` will be called, if
        possible.  The latter will not always be possible if the child is
        terminated by a signal, such as when the parent thread calls
        `child.terminate()` -- `child.stop()` should be used instead.

        The overloaded method MUST return `True` or `False` -- the child will
        continue to work upon `True`, and otherwise (on `False`) begin
        termination.
        '''

        raise NotImplementedError('ru.Thread.work_cb() MUST be overloaded')




# ------------------------------------------------------------------------------
#
# thread-related utility classes and methods
#
class RLock(object):
    """
    This mt.RLock wrapper is supportive of lock debugging.  The only
    semantic difference to mt.RLock is that a lock acquired via the
    'with' statement can be released within the 'with' scope, w/o penalty when
    leaving the locked scope.  This supports up-calling callback semantics, but
    should be used with utter care, and rarely (such as on close()).

    see http://stackoverflow.com/questions/6780613/
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, obj=None):

        self._lock = mt.RLock()


    # --------------------------------------------------------------------------
    #
    def acquire(self):

        self._lock.acquire()


    # --------------------------------------------------------------------------
    #
    def release(self):

        try:
            self._lock.release()

        except RuntimeError as e:
            # lock has been released meanwhile - we allow that
            pass


    # --------------------------------------------------------------------------
    #
    def __enter__(self)                        : self.acquire() 
    def __exit__ (self, type, value, traceback): self.release()



# ------------------------------------------------------------------------------
#
def get_thread_name():

    return mt.current_thread().name


# ------------------------------------------------------------------------------
#
def get_thread_id():

    return mt.current_thread().ident


# ------------------------------------------------------------------------------
#
def gettid():
    """
    Python is not able to give us the native thread ID.  We thus use a syscall
    to do so.  Since this is not portable, we fall back to None in case of error
    (Hi MacOS).
    """
    try:
        import ctypes
        SYS_gettid = 186
        libc = ctypes.cdll.LoadLibrary('libc.so.6')
        return int(libc.syscall(SYS_gettid))
    except:
        return None



# ------------------------------------------------------------------------------
#
def is_main_thread(t=None):

    if t:
        assert(isinstance(t, mt.Thread))
    else:
        t = this_thread()

    return isinstance(t, mt._MainThread)


# ------------------------------------------------------------------------------
#
def is_this_thread(t):
    '''
    check if the given thread (type: threading.Thread) is the same as the
    current thread of execution.
    '''

    assert(isinstance(t, mt.Thread))

    return(t == this_thread())


# ------------------------------------------------------------------------------
#
def main_thread():
    '''
    return a handle to the main thread of execution in this process
    '''

    for t in mt.enumerate():
        if isinstance(t, mt._MainThread):
            return T

    assert(False), 'main thread not found'


# ------------------------------------------------------------------------------
#
def this_thread():
    '''
    return a handle to the current thread of execution
    '''

    return mt.current_thread()


# ------------------------------------------------------------------------------
#
_signal_lock = mt.Lock()
_signal_sent = dict()
def cancel_main_thread(signame=None, once=False):
    """
    This method will call thread.interrupt_main from any calling subthread.
    That will cause a 'KeyboardInterrupt' exception in the main thread.  This
    can be excepted via `except KeyboardInterrupt`
    
    The main thread MUST NOT have a SIGINT signal handler installed (other than
    the default handler or SIGIGN), otherwise this call will cause an exception
    in the core python signal handling (see http://bugs.python.org/issue23395).

    The thread will exit after this, via sys.exit(0), and can then be joined
    from the main thread.

    When being called *from* the main thread, no interrupt will be generated,
    but sys.exit() will still be called.  This can be excepted in the code via 
    `except SystemExit`.

    Another way to avoid the SIGINT problem is to send a different signal to the
    main thread.  We do so if `signal` is specified.

    After sending the signal, any sub-thread will call sys.exit(), and thus
    finish.  We leave it to the main thread though to decide if it will exit at
    this point or not.  Either way, it will have to handle the signal first.

    If `once` is set to `True`, we will send the given signal at most once.
    This will mitigate races between multiple error causes, specifically during
    finalization.
    """

    global _signal_lock
    global _signal_sent

    if signame: signal = get_signal_by_name(signame)
    else      : signal = None

    with _signal_lock:

        if once:
            if signal in _signal_sent:
                # don't signal again
                return

        if signal:
            # send the given signal to the process to which this thread belongs
            os.kill(os.getpid(), signal)
        else:
            # this sends a SIGINT, resulting in a KeyboardInterrupt.
            # NOTE: see http://bugs.python.org/issue23395 for problems on using
            #       SIGINT in combination with signal handlers!
            try:
                thread.interrupt_main()
            except TypeError:
                # this is known to be a side effect of `thread.interrup_main()`
                pass

        # record the signal sending
        _signal_sent[signal] = True

    # the sub thread will at this point also exit.
    if not is_main_thread():
        sys.exit()


# ------------------------------------------------------------------------------
# this is the counterpart for the `cancel_main_thread` method above: any main
# thread should call `set_cancellation_handler`, which will set a signal handler
# for `SIGUSR2`, and upon catching it will raise a `KeyboardInterrupt`
# exception, which can be caught by any interested library or application.
#
# RU claims ownership of `SIGUSR2` -- it will complain if any other signal
# handler is installed for that signal.  
#
# This method can safely be called multiple times.  This method can be called
# from threads, but it will have no effect then (Python allows signal handlers
# to only be installed in the main thread)
#
# FIXME: `cancel_main_thread()` supports arbitrary signals --
#        `set_cancellation_handler()` should, too.
#
def _sigusr2_handler(signum, frame):
    print 'caught sigusr2 (%s)' % os.getpid()
    # we only want to get this exception once, so we unset the signal handler
    # before we raise it
    signal.signal(signal.SIGUSR2, signal.SIG_IGN)
    raise KeyboardInterrupt('sigusr2')

def set_cancellation_handler():

    # check if any handler exists
    old_handler = signal.getsignal(signal.SIGUSR2)
    if  old_handler not in [signal.SIG_DFL, signal.SIG_IGN, None] and \
        old_handler != _sigusr2_handler:
        raise RuntimeError('handler for SIGUSR2 is already present')

    try:
        signal.signal(signal.SIGUSR2, _sigusr2_handler)
    except ValueError:
        # this fails in subthreads - ignore
        pass


def unset_cancellation_handler():

    try:
        signal.signal(signal.SIGUSR2, signal.SIG_IGN)
    except ValueError:
        # this fails in subthreads - ignore
        pass


# ------------------------------------------------------------------------------
#
def get_signal_by_name(signame):
    """
    Translate a signal name into the respective signal number.  If `signame` is
    not found to be a valid signal name, this method will raise a `KeyError`
    exception.  Lookup is case insensitive.
    """

    table = {'abrt'    : signal.SIGABRT,
             'alrm'    : signal.SIGALRM,
             'bus'     : signal.SIGBUS,
             'chld'    : signal.SIGCHLD,
             'cld'     : signal.SIGCLD,
             'cont'    : signal.SIGCONT,
             'fpe'     : signal.SIGFPE,
             'hup'     : signal.SIGHUP,
             'ill'     : signal.SIGILL,
             'int'     : signal.SIGINT,
             'io'      : signal.SIGIO,
             'iot'     : signal.SIGIOT,
             'kill'    : signal.SIGKILL,
             'pipe'    : signal.SIGPIPE,
             'poll'    : signal.SIGPOLL,
             'prof'    : signal.SIGPROF,
             'pwr'     : signal.SIGPWR,
             'quit'    : signal.SIGQUIT,
             'rtmax'   : signal.SIGRTMAX,
             'rtmin'   : signal.SIGRTMIN,
             'segv'    : signal.SIGSEGV,
             'stop'    : signal.SIGSTOP,
             'sys'     : signal.SIGSYS,
             'term'    : signal.SIGTERM,
             'trap'    : signal.SIGTRAP,
             'tstp'    : signal.SIGTSTP,
             'ttin'    : signal.SIGTTIN,
             'ttou'    : signal.SIGTTOU,
             'urg'     : signal.SIGURG,
             'usr1'    : signal.SIGUSR1,
             'usr2'    : signal.SIGUSR2,
             'vtalrm'  : signal.SIGVTALRM,
             'winch'   : signal.SIGWINCH,
             'xcpu'    : signal.SIGXCPU,
             'xfsz'    : signal.SIGXFSZ,
             }
    
    return table[signame.lower()]


# ------------------------------------------------------------------------------
#
class ThreadExit(SystemExit):
    pass

class SignalRaised(SystemExit):

    def __init__(self, msg, signum=None):
        if signum:
            msg = '%s [signal: %s]' % (msg, signum)
        SystemExit.__init__(self, msg)


# ------------------------------------------------------------------------------
#
def raise_in_thread(e=None, tname=None, tident=None):
    """
    This method uses an internal Python function to inject an exception 'e' 
    into any given thread.  That thread can be specified by its name ('tname')
    or thread id ('tid').  If not specified, the exception is sent to the
    MainThread.

    The target thread will receive the exception with some delay.  More
    specifically, it needs to call up to 100 op codes before the exception 
    is evaluated and raised.  The thread interruption can thus be delayed
    significantly, like when the thread sleeps.

    The default exception raised is 'radical.utils.ThreadExit' which inherits
    from 'SystemExit'.

    NOTE: this is not reliable: the exception is not raised immediately, but is
          *scheduled* for raising at some point, ie. in general after about 100
          opcodes (`sys.getcheckinterval()`).  Depending on when exactly the 
          exception is finally raised, the interpreter might silently swallow
          it, if that happens in a generic try/except clause.  Those exist in
          the Python core, even if discouraged by some PEP or the other.

          See https://bugs.python.org/issue1779233


    NOTE: we can only raise exception *types*, not exception *instances*
          See https://bugs.python.org/issue1538556

    Example:

        # ----------------------------------------------------------------------
        def sub():
            time.sleep(0.1)
            ru.raise_in_thread()

        try:
            t = mt.Thread(target=sub)
            t.start()

            while True:
                time.sleep(0.01)

        except ru.ThreadExit:  print 'thread exit'
        except Exception as e: print 'except: %s' % e
        except SystemExit:     print 'exit'
        else:                  print 'unexcepted'
        finally:               t.join()
        # ----------------------------------------------------------------------
    """

    if not tident:
        if not tname:
            tname = 'MainThread'

        for th in mt.enumerate():
            if tname  == th.name:
                tident = th.ident
                break

    if not tident:
        raise ValueError('no target thread given/found')

    if not e:
        e = ThreadExit

    self_thread = mt.current_thread()
    if self_thread.ident == tident:
        # if we are in the target thread, we simply raise the exception.  
        # This specifically also applies to the main thread.
        # Alas, we don't have a decent message to use...
        raise e('raise_in_thread')

    else:
        # otherwise we inject the exception into the main thread's async 
        # exception scheduler
        import ctypes
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tident),
                                                   ctypes.py_object(e))


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


''' Provides utilities to work with configuration files and query strings.  '''

from copy import deepcopy
import ConfigParser, os

class ConfigFileReader(object):

    def __init__(self, filenames):
        config = ConfigParser.RawConfigParser()
        config.read(filenames)

        self.config_dict = dict()
        
        for section in config.sections():
            self.config_dict[section] = dict()
            for option in config.options(section):
                self.config_dict[section][option] = config.get(section, option)

    def get_config_dict(self):
        return self.config_dict

class BaseConfig(object):
    
    def __init__(self):
        self._query_string = ""
        self._dictionary = {}

    def _re_eval(self, query_string=None):
        if query_string is not None:
            # initial eval from query string
            self._query_string = query_string
            for param in self._query_string.split('&'):
                try:
                    (key, value) = param.split("=")
                    self._dictionary[key] = value
                except ValueError:
                    if param != '':
                        self._dictionary[param] = ""
        else:
            # re-genereate query string
            new_query_string = ''
            for (key, value) in self._dictionary.iteritems():
                if value is None:
                    new_query_string += "&%s" % key
                else:
                    new_query_string += "&%s=%s" % (key, value)
            if new_query_string.startswith("&"):
                new_query_string = new_query_string[1:]
            self._query_string = new_query_string
            

    def as_query_string(self, subsection_name=None):
        ''' Return the configuration (or a subsection of it) as a URL 
            query string.
        '''
        self._re_eval()
        return deepcopy(self._query_string)

    def as_config_file(self, subsection_name=None):
        ''' Return the configuration (or a subsection of it) as a  
            configuration file string.
        '''
        self._re_eval()

    def as_dict(self, subsection_name=None):
        ''' Return the configuration (or a subsection of it) as a
            Python dictionary.
        '''
        self._re_eval()
        return deepcopy(self._dictionary)

    def set_option(self, key, value=None):
        self._dictionary[key] = value
        self._re_eval

    def remove_option(self, key):
        del self._dictionary[key]
        self._re_eval

class ConfigFile(BaseConfig):
    ''' Read and parse a configuration from a local file. 
    '''
    def __init__(self, file_path):
        Configuration.__init__(self)

class ConfigQuery(BaseConfig):
    ''' Parse a configuration from a URL query string.
    '''
    def __init__(self, query_string):
        Configuration.__init__(self)
        self._re_eval(query_string=query_string)


def _test_():
    cq = ConfigQuery("WhenToTransferOutput=ON_EXIT&should_transfer_files=YES&notifications=Always")
    print cq.as_dict()
    cq.set_option('VERBOSE')
    print cq.as_query_string()
    cq.set_option('VERBOSE',False)
    print cq.as_query_string()
    cq.remove_option('VERBOSE')
    print cq.as_dict()







import os
import sys
import time
import glob
import regex
import shlex
import signal
import socket
import importlib
import netifaces
import threading

import subprocess as sp
import url as ruu

# ------------------------------------------------------------------------------
#
def split_dburl(dburl, default_dburl=None):
    """
    we split the url into the base mongodb URL, and the path element, whose
    first element is the database name, and the remainder is interpreted as
    collection id.
    """

    # if the given URL does not contain schema nor host, the default URL is used
    # as base, and the given URL string is appended to the path element.

    url = ruu.Url(dburl)

    if not url.schema and not url.host:
        url      = ruu.Url(default_dburl)
        url.path = dburl

    # NOTE: add other data base schemes here...
    if 'mongodb' not in url.schema.split('+'):
        raise ValueError("url must be a 'mongodb://' or 'mongodb+ssl://' url, not '%s'" % dburl)

    host = url.host
    port = url.port
    path = url.path
    user = url.username
    pwd  = url.password
    ssl  = False

    if 'ssl' in url.schema.split('+'):
        ssl = True
        url.schema = 'mongodb'

    if not host:
        host = 'localhost'

    if  path.startswith('/'):
        path = path[1:]
    path_elems = path.split('/')

    dbname = None
    cname  = None
    pname  = None

    if  len(path_elems)  >  0:
        dbname = path_elems[0]

    if  len(path_elems)  >  1:
        dbname = path_elems[0]
        cname  = path_elems[1]

    if  len(path_elems)  >  2:
        dbname = path_elems[0]
        cname  = path_elems[1]
        pname  = '.'.join(path_elems[2:])

    if  dbname == '.':
        dbname = None

    return [host, port, dbname, cname, pname, user, pwd, ssl]


# ------------------------------------------------------------------------------
#
def mongodb_connect(dburl, default_dburl=None):
    """
    connect to the given mongodb, perform auth for the database (if a database
    was given).
    """

    try:
        import pymongo
    except ImportError:
        msg  = " \n\npymongo is not available -- install radical.utils with: \n\n"
        msg += "  (1) pip install --upgrade -e '.[pymongo]'\n"
        msg += "  (2) pip install --upgrade    'radical.utils[pymongo]'\n\n"
        msg += "to resolve that dependency (or install pymongo manually).\n"
        msg += "The first version will work for local installation, \n"
        msg += "the second one for installation from pypi.\n\n"
        raise ImportError(msg)

    [host, port, dbname, cname, pname, user, pwd, ssl] = split_dburl(dburl, default_dburl)

    mongo = pymongo.MongoClient(host=host, port=port, ssl=ssl)
    db    = None

    if  dbname:
        db = mongo[dbname]

        if  user and pwd:
            db.authenticate(user, pwd)

    else:

        # if no DB is given, we try to auth against all databases.
        for dbname in mongo.database_names():
            try:
                mongo[dbname].authenticate(user, pwd)
            except Exception as e:
                pass


    return mongo, db, dbname, cname, pname


# ------------------------------------------------------------------------------
#
def parse_file_staging_directives(directives):
    """
    staging directives

       [local_path] [operator] [remote_path]

    local path:
        * interpreted as relative to the application's working directory
        * must point to local storage (localhost)

    remote path
        * interpreted as relative to the job's working directory

    operator :
        * >  : stage to remote target, overwrite if exists
        * >> : stage to remote target, append    if exists
        * <  : stage to local  target, overwrite if exists
        * << : stage to local  target, append    if exists

    This method returns a tuple [src, tgt, op] for each given directive.  This
    parsing is backward compatible with the simple staging directives used
    previously -- any strings which do not contain staging operators will be
    interpreted as simple paths (identical for src and tgt), operation is set to
    '=', which must be interpreted in the caller context.
    """

    bulk = True
    if  not isinstance(directives, list):
        bulk       = False
        directives = [directives]

    ret = list()

    for directive in directives:

        if  not isinstance(directive, basestring):
            raise TypeError("file staging directives muct by of type string, "
                            "not %s" % type(directive))

        rs = regex.ReString(directive)

        if  rs // '^(?P<one>.+?)\s*(?P<op><|<<|>|>>)\s*(?P<two>.+)$':
            res = rs.get()
            ret.append([res['one'], res['two'], res['op']])

        else:
            ret.append([directive, directive, '='])

    if  bulk: return ret
    else    : return ret[0]


# ------------------------------------------------------------------------------
#
def time_stamp(spec):

    if  isinstance(spec, int)   or \
        isinstance(spec, float)    :

        import datetime
        return datetime.datetime.utcfromtimestamp(spec)

    return spec


# ------------------------------------------------------------------------------
#
def time_diff(dt_abs, dt_stamp):
    """
    return the time difference bewteen  two datetime
    objects in seconds (incl. fractions).  Exceptions (like on improper data
    types) fall through.
    """

    delta = dt_stamp - dt_abs

    # make it easy to use seconds since epoch instead of datetime objects
    if  isinstance(delta, int)   or \
        isinstance(delta, float)    :
        return delta

    import datetime
    if  not isinstance(delta, datetime.timedelta):
        raise TypeError("difference between '%s' and '%s' is not a .timedelta"
                     % (type(dt_abs), type(dt_stamp)))

    # get seconds as float
    seconds = delta.seconds + delta.microseconds / 1E6
    return seconds


# ------------------------------------------------------------------------------
#
def all_pairs(iterable, n):
    """
    [ABCD] -> [AB], [AC], [AD], [BC], [BD], [CD]
    """

    import itertools
    return list(itertools.combinations(iterable, n))


# ------------------------------------------------------------------------------
#
def cluster_list(iterable, n):
    """
    s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...
    """

    from itertools import izip
    return izip(*[iter(iterable)] * n)


# ------------------------------------------------------------------------------
# From https://docs.python.org/release/2.3.5/lib/itertools-example.html
#
def window(seq, n=2):
    """
    Returns a sliding window (of width n) over data from the iterable"
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    """

    from itertools import islice

    it = iter(seq)
    result = tuple(islice(it, n))

    if len(result) == n:
        yield result

    for elem in it:
        result = result[1:] + (elem,)
        yield result


# ------------------------------------------------------------------------------
#
def round_to_base(value, base=1):
    """
    This method expects an integer or float value, and will round it to any
    given integer base.  For example:

      1.5, 2 -> 2
      3.5, 2 -> 4
      4.5, 2 -> 4

      11.5, 20 -> 20
      23.5, 20 -> 20
      34.5, 20 -> 40

    The default base is '1'.
    """

    return int(base * round(float(value) / base))


# ------------------------------------------------------------------------------
#
def round_upper_bound(value):
    """
    This method expects an integer or float value, and will return an integer upper
    bound suitable for example to define plot ranges.  The upper bound is the
    smallest value larger than the input value which is a multiple of 1, 2 or
    5 times the order of magnitude (10**x) of the value.
    """

    bound = 0
    order = 0
    check = [1, 2, 5]

    while True:

        for c in check:

            bound = c * (10**order)

            if value < bound:
                return bound

        order += 1


# ------------------------------------------------------------------------------
#
def islist(thing):
    """
    return True if a thing is a list thing, False otherwise
    """

    return isinstance(thing, list)


# ------------------------------------------------------------------------------
#
def tolist(thing):
    """
    return a non-list thing into a list thing
    """

    if islist(thing):
        return thing
    return [thing]


# ------------------------------------------------------------------------------
#
# to keep RU 2.6 compatible, we provide import_module which works around some
# quirks of __import__ when being used with dotted names. This is what the
# python docs recommend to use.  This basically steps down the module path and
# loads the respective submodule until arriving at the target.
#
def import_module(name):

    mod = __import__(name)
    for s in name.split('.')[1:]:
        mod = getattr(mod, s)
    return mod


# ------------------------------------------------------------------------------
#
_hostname = None
def get_hostname():
    """
    Look up the hostname
    """

    global _hostname
    if not _hostname:

        if socket.gethostname().find('.') >= 0:
            _hostname = socket.gethostname()

        else:
            _hostname = socket.gethostbyaddr(socket.gethostname())[0]

    return _hostname


# ------------------------------------------------------------------------------
#
_hostip = None
def get_hostip(req=None, logger=None):
    """
    Look up the ip number for a given requested interface name.
    If interface is not given, do some magic.
    """

    AF_INET = netifaces.AF_INET

    # We create a ordered preference list, consisting of:
    #   - given arglist
    #   - white list (hardcoded preferred interfaces)
    #   - black_list (hardcoded unfavorable interfaces)
    #   - all others (whatever is not in the above)
    # Then this list is traversed, we check if the interface exists and has an
    # IP address.  The first match is used.

    if req: 
        if not isinstance(req, list):
            req = [req]
    else:
        req = []

    white_list = [
            'ipogif0', # Cray's
            'br0',     # SuperMIC
            'eth0',    # desktops etc.
            'wlan0'    # laptops etc.
            ]

    black_list = [
            'lo',      # takes the 'inter' out of the 'net'
            'sit0'     # ?
            ]

    all  = netifaces.interfaces()
    rest = [iface for iface in all \
                   if iface not in req and \
                      iface not in white_list and \
                      iface not in black_list]

    preflist = req + white_list + black_list + rest

    for iface in preflist:

        if iface not in all:
            if logger:
                logger.debug('check iface %s: does not exist', iface)
            continue

        info = netifaces.ifaddresses(iface)
        if AF_INET not in info:
            if logger:
                logger.debug('check iface %s: no information', iface)
            continue

        if not len(info[AF_INET]):
            if logger:
                logger.debug('check iface %s: insufficient information', iface)
            continue

        if not info[AF_INET][0].get('addr'):
            if logger:
                logger.debug('check iface %s: disconnected', iface)
            continue

      
        ip = info[AF_INET][0].get('addr')
        if logger:
            logger.debug('check iface %s: ip is %s', iface, ip)

        if ip:
           _hostip = ip
           return ip

    raise RuntimeError('could not determine ip on %s' % preflist)


# ------------------------------------------------------------------------------
#
def watch_condition(cond, target=None, timeout=None, interval=0.1):
    """
    Watch a given condition (a callable) until it returns the target value, and
    return that value.  Stop watching on timeout, in that case return None.  The
    condition is tested approximately every 'interval' seconds.
    """
    
    start = time.time()
    while True:
        ret = cond()
        if ret == target:
            return ret
        if timeout and time.time() > start+timeout:
            return None
        time.sleep(interval)


# ------------------------------------------------------------------------------
#
def name2env(name):
    """
    convert a name of the for 'radical.pilot' to an env vare base named
    'RADICAL_PILOT'.
    """
    
    return name.replace('.', '_').upper()


# ------------------------------------------------------------------------------
#
def stack():

    ret = {'sys'     : {'python'     : sys.version.split()[0],
                        'pythonpath' : os.environ.get('PYTHONPATH',  ''),
                        'virtualenv' : os.environ.get('VIRTUAL_ENV', '') or os.environ.get('CONDA_DEFAULT_ENV','')}, 
           'radical' : dict()
          }


    modules = ['radical.utils', 
               'radical.saga', 
               'saga', 
               'radical.pilot', 
               'radical.entk',
               'radical.analytics']

    for mod in modules:
        try:
            tmp = importlib.import_module(mod)
            ret['radical'][mod] = tmp.version_detail
        except:
            pass

    return ret
    

# ------------------------------------------------------------------------------
#
def get_size(obj, seen=None, strict=False):

    size   = sys.getsizeof(obj)
    obj_id = id(obj)

    if strict:
        # perform recursion checks
        if seen is None:
            seen = set()
        if obj_id in seen:
            return 0
        seen.add(obj_id)

    if isinstance(obj, dict):
        size += sum([get_size(v, seen, strict) for v in obj.values()])
        size += sum([get_size(k, seen, strict) for k in obj.keys()])

    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen, strict)

    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen, strict) for i in obj])

    return size


# ------------------------------------------------------------------------------
#
def dockerized():

    if os.path.exists('/.dockerenv'):
        return True
    return False


# ------------------------------------------------------------------------------
#
def sh_callout(cmd, shell=False):
    '''
    call a shell command, return `[stdout, stderr, retval]`.
    '''

    # convert string into arg list if needed
    if not shell and isinstance(cmd, basestring):
        cmd = shlex.split(cmd)

    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=shell)
    stdout, stderr = p.communicate()
    return stdout, stderr, p.returncode


# ------------------------------------------------------------------------------
#
def get_radical_base(module=None):
    '''
    Several parts of the RCT stack store state on the file system.  This should
    usually be under `$HOME/.radical` - but that location is not always
    available or desireable.  We interpret the env variable `RADICAL_BASE_DIR`,
    and fall back to `pwd` if neither that nor `$HOME` exists.

    The optional `module` parameter will result in the respective subdir name to
    be appended.  The resulting dir is created (if it does not exist), and the
    name is returned.
    '''


    base = os.environ.get("RADICAL_BASE_DIR")

    if not base or not os.path.isdir(base):
        base  = os.environ.get("HOME")

    if not base or not os.path.isdir(base):
        base  = os.environ.get("PWD")

    if not base or not os.path.isdir(base):
        base  = os.getcwd()

    if module: base += '/.radical/%s/' % module
    else     : base += '/.radical/'

    if not os.path.isdir(base):
        os.makedirs(base)

    return base


# ------------------------------------------------------------------------------




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_singleton():
    '''
    Test if the logger behaves like a singleton
    '''

    # make sure singleton works
    assert ru.get_logger('radical.utils') == ru.get_logger('radical.utils')


# ------------------------------------------------------------------------------
#
def test_logger():
    '''
    Print out some messages with different log levels
    '''

    tmp = ru.get_logger('engine')
    tmp = ru.get_logger('engine')
    tmp.setLevel('DEBUG')

    tmp.debug('debug')
    tmp.info('info')
    tmp.warn('warning')
    tmp.error('error')
    tmp.fatal('fatal')


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


''' 
Unit tests for ru.Process()
'''

import os
import sys
import time

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_process_basic():
    '''
    start a 'sleep 0.2', and expect this to finish within 0.x seconds
    '''

    class P(ru.Process):
        def __init__(self):
            return ru.Process.__init__(self, 'ru.test')

        def work_cb(self):
            time.sleep(0.2)
            return False

    p = P()    ; t1 = time.time()
    p.start()  ; t2 = time.time()
    p.join(10) ; t3 = time.time()

    assert(t2-t1 > 0.0), t2-t1
    assert(t2-t1 < 0.2), t2-t1  # process startup should be quick
    assert(t3-t1 > 0.2), t3-t1  # expect exactly one work iteration
    assert(t3-t2 < 0.5), t3-t2


# ------------------------------------------------------------------------------
#
def test_process_autostart():
    '''
    start the child process on __init__()
    '''

    class P(ru.Process):
        def __init__(self):

            self._initalize_common = False
            self._initalize_parent = False
            self._initalize_child  = False

            self._finalize_common  = False
            self._finalize_parent  = False
            self._finalize_child   = False

            self._work_done        = False

            ru.Process.__init__(self, 'ru.test')

            self.start()

            assert(self._initialize_common), 'no initialize common'
            assert(self._initialize_parent), 'no initialize parent'

            self.join()  # wait until work is done
            self.stop()  # make sure finalization happens

            assert(self._finalize_common), 'no finalize common'
            assert(self._finalize_parent), 'no finalize parent'

        def ru_initialize_common(self): self._initialize_common = True
        def ru_initialize_parent(self): self._initialize_parent = True
        def ru_initialize_child (self): self._initialize_child  = True

        def ru_finalize_common(self)  : self._finalize_common   = True
        def ru_finalize_parent(self)  : self._finalize_parent   = True
        def ru_finalize_child (self)  : self._finalize_child    = True

        def work_cb(self):
            assert(self._initialize_common), 'no initialize common'
            assert(self._initialize_child),  'no initialize child'
            self._work_done = True
            return False  # only run once

    p = P()


# ------------------------------------------------------------------------------
#
def test_process_init_fail():
    '''
    make sure the parent gets notified on failing init
    '''

    class P(ru.Process):
        def __init__(self):
            return ru.Process.__init__(self, 'ru.test')
        def ru_initialize_child(self):
            raise RuntimeError('oops init')
        def work_cb(self):
            time.sleep(0.1)
            return True

    try:
        p = P()
        p.start()
    except RuntimeError as e:
        assert('oops init' in str(e)), str(e)
    else:
        assert(False), 'missing exception'

    assert(not p.is_alive())


# ------------------------------------------------------------------------------
#
def test_process_final_fail():
    '''
    make sure the parent gets notified on failing finalize
    '''

    class P(ru.Process):
        def __init__(self):
            return ru.Process.__init__(self, 'ru.test')
        def ru_initialize_child(self):
            self.i = 0
        def work_cb(self):
            self.i += 1
            if self.i == 5:
                time.sleep(0.1)
                return False
            return True
        def ru_finalize_child(self):
            raise RuntimeError('oops final')

    try:
        p = P()
        p.start()
        p.stop()
    except Exception as e:
        print 'excepted: %s' %e
        assert('oops final' in str(e)), str(e)
    else:
        pass
      # assert(False), 'missing exception'  # FIXME

    assert(not p.is_alive())


# ------------------------------------------------------------------------------
#
def test_process_parent_fail():
    '''
    make sure the child dies when the parent dies
    '''

    class Parent(ru.Process):

        def __init__(self):
            ru.Process.__init__(self, name='ru.test')

        def ru_initialize_child(self):
            self._c = Child()
            self._c.start()
            assert(self._c.is_alive())

        def work_cb(self):
            sys.exit()  # parent dies

        def ru_finalize_child(self):
          # # below is what's needed for *clean* termination
          # self._c.stop()
            pass


    class Child(ru.Process):

        def __init__(self):
            with open('/tmp/c_pid.%d' % os.getuid(), 'w') as f:
                f.write(str(os.getpid()))
            ru.Process.__init__(self, name='ru.test.child')

        def work_cb(self):
            return True


    p = Parent()
    p.start()
    with open('/tmp/c_pid.%d' % os.getuid(), 'r') as f:
        c_pid = int(f.read().strip())
    os.unlink('/tmp/c_pid.%d' % os.getuid())
    os.kill(c_pid, 9)

    # leave some time for child to die
    time.sleep(1.01)
    try:
        os.kill(p.pid, 0)
    except OSError as e:
        pass  # child is gone
    else:
        pass
      # assert(False), 'child not gone'  # FIXME

    assert(not p.is_alive())


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    N = 1

    test_process_autostart()
    for i in range(N):
        test_process_final_fail()
        print '.',
        test_process_init_fail()
        print '.',
        test_process_parent_fail()
        print '.',
        test_process_basic()
        print '.',
        print i

    sys.exit()


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys

import singleton as rs
import read_json as rj


# ------------------------------------------------------------------------------
#
_test_config = {}


# ------------------------------------------------------------------------------
#
def get_test_config () :

    return _test_config


# ------------------------------------------------------------------------------
#
class TestConfig (dict) :

    def __init__ (self, cfg_file, cfg_name=None) :

        config = rj.read_json (cfg_file)

        if  cfg_name and cfg_name in config :
            self._cfg = config[cfg_name]
        else :
            self._cfg = config

        global _test_config
        _test_config = self

        dict.__init__ (self, self._cfg)

        self.changed_values = {}
        self.__initialized  = True

    def __setitem__(self, key, value):
        self.changed_values[key] = value
        super(TestConfig, self).__setitem__(key, value)

    def __getattr__(self, item):
        """Maps values to attributes.
        Only called if there *isn't* an attribute with this name
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            return None

    def __setattr__(self, item, value):
        """Maps attributes to values.
        Only if we are initialised
        """
        if not self.__dict__.has_key('_D__initialized'):  # this test allows attributes to be set in the __init__ method
            return dict.__setattr__(self, item, value)
        elif self.__dict__.has_key(item):       # any normal attributes are handled normally
            dict.__setattr__(self, item, value)
        else:
            self.__setitem__(item, value)


# ------------------------------------------------------------------------------
#
class Testing (object) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, modname, basedir) :
        """
        The given basedir is assumed to be `tests/` beneath the module tree to
        be tested.  We further assume a directory `tests/unittests/` to exist
        and to contain the actual set of test scripts.
        """

        self._testdir = "%s/unittests/" % os.path.dirname (os.path.realpath (basedir))

        try:
            module = __import__ (modname)
        
        except Exception, e:
            srcdir = "%s/../" % os.path.dirname (os.path.realpath (basedir))
            sys.path.insert (0, os.path.abspath(srcdir))
            module = __import__ (modname)


        print "______________________________________________________________________"
        print "Using %s from: %s" % (modname, module.__path__[0])
        print "______________________________________________________________________"

        try :
            import nose
        except ImportError :
            msg  = " \n\nnose is not available -- install radical.utils with: \n\n"
            msg += "  (1) pip install --upgrade -e '.[nose]'\n"
            msg += "  (2) pip install --upgrade    'radical.utils[nose]'\n\n"
            msg += "to resolve that dependency (or install nose manually).\n"
            msg += "The first version will work for local installation, \n"
            msg += "the second one for installation from pypi.\n\n"
            raise ImportError (msg)



    # --------------------------------------------------------------------------
    #
    def run (self) :
        """

        This method runs a set of unit tests, which are organized in a directory
        structure under ``tests/unittests/`` -- each sub-directory in that
        hierarchy represents a test suite to be run.
        
        The unit tests in the individual test suites will have access to the
        same configuration info (parameter test_config), and will use the given
        parameters to set up the test environment.  For example, the saga-python
        api/job/test_service.py unit test will use the following::
        
            import saga.utils.test_config as sutc
            ...
        
            tc = sutc.TestConfig ('test.cfg')
            js = saga.job.Service (tc['js_url'], tc['session'])
        
        The :class:`saga.utils.test_config.TestConfig` class will expose the currently
        active test configuration -- which is activated in the *run_unittests* script
        as follows::
        
            tc = sutc.TestConfig (cfg_name)
        
        Since :class:`saga.utils.test_config.TestConfig` is
        a :class:`saga.utils.singleton.Singleton`, the thusly set state will be shared
        with the test suites as shown.
        
        """
        import nose
        
        tc = get_test_config ()

        results = list()

        # fall back to no subtree structure if no test suites are specified
        if  not 'test_suites' in tc \
        or  not tc['test_suites'] :
            tc['test_suites'] = ['.']

        verbosity_env  = os.getenv('NOSE_VERBOSE', 1)
        verbosity_nose = None

        try    : verbosity_nose = int(verbosity_env)
        except : pass

        try    : verbosity_nose = {'ERROR'   : 1,
                                   'WARNING' : 2,
                                   'INFO'    : 3, 
                                   'DEBUG'   : 4}[verbosity_env]
        except : pass

        if  verbosity_nose == None:
            verbosity_nose =  2

        # run all test suites from the config
        for test_suite in tc['test_suites'] :

            # configure the unit test framework
            config = nose.config.Config ()
        
            config.verbosity  = verbosity_nose
            config.workingDir = self._testdir + '/' + test_suite
            config.stream     = sys.stderr
        
            # and run tests
            print "______________________________________________________________________"
            print "%s" % test_suite
            print "______________________________________________________________________"
            rc = nose.core.run (argv=[sys.argv[0]], config=config)
            results.append(rc)
            print "RC: %s" % rc
            print "______________________________________________________________________"
        
        # if we get a 'false' results it means that something went wrong. in
        # that case we return a non-zero exit code
        
        if False in results:
            return -1
        else:
            return 0


# --------------------------------------------------------------------
#






__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

# we want atfork imported first, specifically before os and logging
from .atfork         import *

# import utility classes
from .object_cache   import ObjectCache
from .plugin_manager import PluginManager
from .singleton      import Singleton
from .process        import Process
from .threads        import Thread, RLock
from .threads        import is_main_thread, is_this_thread, cancel_main_thread
from .threads        import main_thread, this_thread, get_thread_name, gettid
from .threads        import set_cancellation_handler, unset_cancellation_handler
from .threads        import raise_in_thread, ThreadExit, SignalRaised
from .futures        import Future
from .futures        import NEW, RUNNING, DONE, FAILED, CANCELED
from .url            import Url
from .dict_mixin     import DictMixin, dict_merge, dict_stringexpand
from .dict_mixin     import PRESERVE, OVERWRITE
from .lockable       import Lockable
from .registry       import Registry, READONLY, READWRITE
from .regex          import ReString, ReSult
from .reporter       import Reporter
from .benchmark      import Benchmark
from .lease_manager  import LeaseManager
from .daemonize      import Daemon
from .poll           import Poller, POLLIN, POLLOUT, POLLERR, POLLALL
from .poll           import POLLNVAL, POLLPRI, POLLHUP

# import utility methods
from .logger         import *
from .ids            import *
from .read_json      import *
from .tracer         import trace, untrace
from .which          import which
from .debug          import *
from .misc           import *
from .get_version    import get_version
from .algorithms     import *
from .profile        import *

# import decorators
from .timing         import timed_method

# import sub-modules
# from config         import Configuration, Configurable, ConfigOption, getConfig


# ------------------------------------------------------------------------------


import os

_mod_root = os.path.dirname (__file__)

version_short, version_detail, version_base, \
               version_branch, sdist_name,   \
               sdist_path = get_version(_mod_root)
version = version_short

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils as ru

# ------------------------------------------------------------------------------
def test_re_string () :
    """
    Test regex matching
    """

    txt   = ru.ReString ("The quick brown fox jumps over the lazy dog")
    tgt_l = [' qu', 'ick brown fox jumps']
    tgt_d = {'x'  : 'ick brown fox jumps'}
    
    with txt // '(\s.u)(?P<x>.*?j\S+)' as res :
        assert (res)
        assert (len(res) == len(tgt_l))
        assert (res      == tgt_l), "%s != %s" % (str(res), str(tgt_l))
        assert (res[0]   == tgt_l[0])
        assert (res[1]   == tgt_l[1])
        assert (res['x'] == tgt_d['x'])
        assert (res.x    == tgt_d['x'])
        for i, r in enumerate (res) :
            assert (r    == tgt_l[i])
    
    if  txt // '(rabbit)' :
        assert (False)
    
    elif  txt // '((?:\s).{12,15}?(\S+))' :     # support for full Python regex slang
        assert (True)
    
    else :
        assert (False)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_re_string ()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import sys
import os


# ------------------------------------------------------------------------------
#
# Method call parameters/return value type checking decorators.
# (c) 2006-2007, Dmitry Dvoinikov <dmitry@targeted.org>
# Distributed under BSD license.
#
# code adjustements by 'The SAGA Project', 2013
# See also 
# http://code.activestate.com/recipes/577065-type-checking-function-overloading-decorator/
#
#
# Samples:
#
# from typecheck import *
#
# @takes(int, str) # takes   int, str, upon a problem throws TypeError
# @returns(int)    # returns int,      upon a problem throws TypeError
# def foo(i, s): 
#     return i + len(s)
#
# @takes((int, long), by_regex("^[0-9]+$")) # int or long, numerical string
# def foo(i, s, anything):                  # and the third parameter is not checked
#     ...
#
# @takes(int, int, foo = int, bar = optional(int)) # keyword argument foo must be int
# def foo(a, b, **kwargs):                         # bar may be int or missing
#     ...
#
# Note: @takes for positional arguments, @takes for keyword arguments and @returns
# all support the same checker syntax, for example for the following declaration
# 
# @takes(C)
# def foo(x):
#     ...
#
# then C may be one of the simple checkers:
#
# --------- C ---------     ------------- semantics -------------
# typename              ==> ok if x is is an instance of typename
# "typename"            ==> ok if x is is an instance of typename
# with_attr("a", "b")   ==> ok if x has specific attributes
# some_callable         ==> ok if some_callable(x) is True
# one_of(1, "2")        ==> ok if x is one of the literal values
# by_regex("^foo$")     ==> ok if x is a matching basestring
# nothing               ==> ok if x is None
# anything              ==> always ok
#
# simple checkers can further be combined with OR semantics using tuples:
#
# --------- C ---------     ------------- semantics -------------
# (checker1, checker2)      ==> ok if x conforms with either checker
#
# be optional:
#
# --------- C ---------     ------------- semantics -------------
# optional (checker)        ==> ok if x is checker-conformant or None
#
# or nested recursively into one of the following checkers
#
# --------- C ---------     ------------- semantics -------------
# list_of  (checker)        ==> ok if x is a list  of checker-conformant values
# tuple_of (checker)        ==> ok if x is a tuple of checker-conformant values
# set_of   (checker)        ==> ok if x is a set   of checker-conformant values
# dict_of  (key_checker, 
#           value_checker)  ==> ok if x is a dict  mapping key_checker-
#                           conformant keys to value_checker-conformant values
#
# More samples:
#
# class foo (object):
#     @takes ("foo", optional (int)) # foo, maybe int, but foo is yet incomplete
#     def __init__ (self, i = None): # and is thus specified by name
#         ...
#     @takes ("foo", int)            # foo, and int if presents in args,
#     def bar (self, *args):         # if args is empty, the check passes ok
#         ...
#     @takes ("foo")                 
#     @returns (object)              # returns foo which is fine, because
#     def biz (self):                # foo is an object
#         return self
#     @classmethod                   # classmethod's and staticmethod's
#     @takes (type)                  # go same way
#     def baz (cls):
#         ...
#    
# @takes   (int)
# @returns (optional ("int", foo))   # returns either int, foo or NoneType
# def bar (i):                       # "int" (rather than just int) is for fun
#     if i > 0: 
#         return i
#     elif i == 0:
#         return foo()               # otherwise returns NoneType
#
# @takes   (callable)                # built-in functions are treated as predicates
# @returns (lambda x: x == 123)      # and so do user-defined functions or lambdas
# def exec  (f, *args, **kwargs):
#     return f (*args, **kwargs)
#
# assert execute (execute, execute, execute, lambda x: x, 123) == 123
#
# def readable (x):                  # user-defined type-checking predicate
#     return hasattr (x, "read")
#
# anything is an alias for predicate lambda: True,
# nothing is an alias for NoneType, as in:
#
# @takes   (callable, readable, 
#           optional (anything), 
#           optional (int))
# @returns (nothing)
# def foo (f, r, x = None, i = None):
#     ...
#
#                                    # another way of protocol checking
# @takes (with_attr ("read", "write")) 
# def foo (pipe):
#     ...
#
# @takes (list_of (int))             # list of ints
# def foo (x):
#     print x[0]
#
# @takes (tuple_of (callable))       # tuple of callables
# def foo(x):
#     print x[0]()
#
#                                    # dict mapping strs to lists of int
# @takes (dict_of (str, list_of (int))) 
# def foo (x):
#     print sum (x["foo"])
#
# @takes (by_regex ("^[0-9]{1,8}$")) # integer-as-a-string regex
# def foo (x):
#     i = int (x)
#
# @takes (one_of (1, 2))             # must be equal to either one
# def set_version (version):
#     ...
#
# The (3 times longer) source code with self-tests is available from:
# http://www.targeted.org/python/recipes/typecheck.py
#
# ------------------------------------------------------------------------------

__all__ = [ "takes",    "returns",   "optional", "nothing",   
            "anything", "list_of",   "tuple_of", "dict_of",
            "by_regex", "with_attr", "one_of",   "set_of" ]

no_check        = True  # set this to True to turn all checks off
no_return_check = True   # set this to True to turn return value cchecks off

if 'RADICAL_DEBUG_SIG' in os.environ:
    no_check    = False  # set this to True to turn all checks off

# ------------------------------------------------------------------------------

from traceback import extract_stack
from inspect   import getargspec, isclass
from types     import NoneType
from re        import compile as regex
from functools import wraps

# ------------------------------------------------------------------------------

def base_names (C):
    "Returns list of base class names for a given class"
    return [ x.__name__ for x in C.__mro__ ]
    
# ------------------------------------------------------------------------------

def type_name (v):
    "Returns the name of the passed value's type"
    return type (v).__name__

# ------------------------------------------------------------------------------

class Checker (object):

    def __init__ (self, reference):
        self.reference = reference
        self.spectype  = reference

    def check (self, value): # abstract
        pass

    _registered = [] # a list of registered descendant class factories

    @staticmethod
    def create (value): # static factory method
        for f, t in Checker._registered:
            if f (value):
                return t (value)
        else:
            return None

# ------------------------------------------------------------------------------

class TypeChecker (Checker):

    def check (self, value):
        return isinstance (value, self.reference)

Checker._registered.append ((isclass, TypeChecker))
nothing = NoneType

# ------------------------------------------------------------------------------

class StrChecker (Checker):

    def check (self, value):
        value_base_names = base_names (type (value))
        return self.reference in value_base_names or \
               "instance"     in value_base_names
   
Checker._registered.append ((lambda x: isinstance (x, str), StrChecker))

# ------------------------------------------------------------------------------

class TupleChecker (Checker):

    def __init__ (self, reference):
        self.reference = map (Checker.create, reference)
        self.spectype  = reference

    def check (self, value):
        return reduce (lambda r, c: r or c.check (value), self.reference, False)

Checker._registered.append ((lambda x: isinstance (x, tuple) and not
                     filter (lambda y: Checker.create (y) is None, x), 
                     TupleChecker))

optional = lambda *args: args + (NoneType, )

# ------------------------------------------------------------------------------

class CallableChecker (Checker):

    def check (self, value):
        return self.reference (value)

# note that the callable check is the most relaxed of all, therefore it should
# be registered last, after all the more specific cases have been registered

Checker._registered.append ((callable, CallableChecker))

anything = lambda *args: True

# ------------------------------------------------------------------------------

class ListOfChecker (Checker):

    def __init__ (self, reference):
        self.reference = Checker.create (reference)

    def check (self, value):
        return isinstance (value, list) and \
               not filter (lambda e: not self.reference.check (e), value)

list_of = lambda *args: ListOfChecker (*args).check

# ------------------------------------------------------------------------------

class TupleOfChecker (Checker):

    def __init__ (self, reference):
        self.reference = Checker.create (reference)

    def check (self, value):
        return isinstance (value, tuple) and \
               not filter (lambda e: not self.reference.check (e), value)

tuple_of = lambda *args: TupleOfChecker (*args).check

# ------------------------------------------------------------------------------

class SetOfChecker (Checker):

    def __init__ (self, reference):
        self.reference = Checker.create (reference)

    def check (self, value):
        return isinstance (value, set) and \
               not filter (lambda e: not self.reference.check (e), value)

set_of = lambda *args: SetOfChecker (*args).check

# ------------------------------------------------------------------------------

class DictOfChecker (Checker):

    def __init__ (self, key_reference, value_reference):
        self.key_reference   = Checker.create (key_reference)
        self.value_reference = Checker.create (value_reference)

    def check (self, value):
        return isinstance (value, dict) and \
               not filter (lambda e: not self.key_reference.check   (e), value.iterkeys  ()) and \
               not filter (lambda e: not self.value_reference.check (e), value.itervalues())

dict_of = lambda *args: DictOfChecker (*args).check

# ------------------------------------------------------------------------------

class RegexChecker (Checker):

    def __init__(self, reference):
        self.reference = regex (reference)

    def check (self, value):
        return isinstance (value, basestring) and self.reference.match (value)

by_regex = lambda *args: RegexChecker (*args).check

# ------------------------------------------------------------------------------

class AttrChecker (Checker):

    def __init__ (self, *attrs):
        self.attrs = attrs

    def check (self, value):
        return reduce (lambda r, c: r and c, 
                  map (lambda a: hasattr (value, a), self.attrs), True)

with_attr = lambda *args: AttrChecker (*args).check

# ------------------------------------------------------------------------------

class OneOfChecker (Checker):

    def __init__ (self, *values):
        self.values = values

    def check (self, value):
        return value in self.values

one_of = lambda *args: OneOfChecker (*args).check


def create_return_exception (method, spectype, result) :

    stack = extract_stack ()
    for f in stack : 
        if  'utils/signatures.py' in f[0] :
            break
        frame = f

    msg   = "\nSignature Mismatch\n"
    msg  += "  in function   : %s\n"       % (frame[2])
    msg  += "  in file       : %s +%s\n"   % (frame[0], frame[1])
    msg  += "  on line       : %s\n"       % (frame[3])
    msg  += "  method        : %s\n"       % (method.__name__)
    msg  += "  returned type : %s\n"       % (type(result).__name__)
    msg  += "  instead  of   : %s\n"       % (spectype.__name__)
    msg  += "  This is an internal error!"

    return TypeError (msg)


def create_type_exception (method, arg0, i, arg, spectype, kwname="") :

    narg = 0
    
    if  arg0 and isinstance (arg0, object) :
        narg = i
    else :
        narg = i+1
    
    stack = extract_stack ()
    for f in stack : 
        if  'utils/signatures.py' in f[0] :
            break
        frame = f

    if isinstance(spectype,tuple) or \
       isinstance(spectype,list )  :
        expected = list()
        for t in spectype :
            if isinstance(t, type):
                expected.append (t.__name__)
            else:
                expected.append (str(t))

    elif isinstance(spectype,type):
        expected = spectype.__name__

    else:
        expected = str(spectype)

    msg   = "\nSignature Mismatch\n"
    msg  += "  in function        : %s\n"       % (frame[2])
    msg  += "  in file            : %s +%s\n"   % (frame[0], frame[1])
    msg  += "  on line            : %s\n"       % (frame[3])
    msg  += "  method             : %s\n"       % (method.__name__)
    if  not kwname :
        msg  += "  argument           : #%s\n"  % (narg)
    else :
        msg  += "  parameter          : %s"     % (kwname)
    msg  += "  has incorrect type : %s\n"       % (type (arg).__name__)
    msg  += "  instead of         : %s"         % (str(expected))

    return TypeError (msg)


# ------------------------------------------------------------------------------

def takes (*args, **kwargs):
    "Method signature checking decorator"

    if no_check: # no type checking is performed, return decorated method itself

        def takes_proxy (method):
            return method

    else:

        # convert decorator arguments into a list of checkers

        checkers = []
        for i, arg in enumerate (args):
            checker = Checker.create (arg)
            if checker is None:
                raise TypeError ("@takes decorator got parameter %d of unsupported "
                                 "type %s" % (i + 1, type_name(arg)))
            checkers.append (checker)

        kwcheckers = {}
        for kwname, kwarg in kwargs.iteritems ():
            checker = Checker.create (kwarg)
            if checker is None:
                raise TypeError ("@takes decorator got parameter %s of unsupported "
                                 "type %s" % (kwname, type_name (kwarg)))
            kwcheckers[kwname] = checker


        def takes_proxy (method):
            
            method_args, method_defaults = getargspec (method)[0::3]

            @wraps(method)
            def signature_check (*pargs, **pkwargs):
    
                # append the default parameters

                if  method_defaults is not None and len (method_defaults) > 0 \
                    and len (method_args) - len (method_defaults) <= len (pargs) < len (method_args):
                    pargs += method_defaults[len (pargs) - len (method_args):]

                # check the types of the actual call parameters

                for i, (arg, checker) in enumerate (zip (pargs, checkers)):
                    if  not checker.check (arg):
                      # print 'Checker.spectype %s' % checker.spectype
                        raise (create_type_exception (method, pargs[0], i,
                                                      arg, checker.spectype))

                for kwname, checker in kwcheckers.iteritems ():
                    if  not checker.check (pkwargs.get (kwname, None)):
                      # print 'checker.spectype %s' % checker.spectype
                        raise (create_type_exception (method, pargs[0], i,
                                                      arg, checker.spectype, kwname))

                try :
                    return method(*pargs, **pkwargs)
                except :
                    # remove signature decorator from exception call stack
                    et, ei, tb = sys.exc_info()
                    raise et, ei, tb.tb_next

            signature_check.__name__ = method.__name__
            return signature_check
    
    return takes_proxy

# ------------------------------------------------------------------------------

def returns (sometype):
    "Return type checking decorator"

    if no_check: # no type checking is performed, return decorated method itself

        def returns_proxy (method):
            return method

    else:

        # convert decorator argument into a checker
        
        checker = Checker.create(sometype)
        if checker is None:
            raise TypeError ("@returns decorator got parameter of unsupported "
                             "type %s" % type_name (sometype))

        def returns_proxy (method):
            
            @wraps(method)
            def signature_check (*args, **kwargs):
                
                try :
                    result = method (*args, **kwargs)
                except :
                    # remove signature decorator from exception call stack
                    et, ei, tb = sys.exc_info()
                    raise et, ei, tb.tb_next
                
                if  not checker.check (result):
                    
                    if not  no_return_check:
                      # print 'Checker.spectype %s' % checker.spectype
                        raise (create_return_exception (method, checker.spectype, result))

                return result
    
            signature_check.__name__ = method.__name__
            return signature_check
        
    return returns_proxy


# ------------------------------------------------------------------------------






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys
import math
import time
import socket
import threading
import traceback

import threads as rut
import testing as rutest


# --------------------------------------------------------------------
#
class Benchmark (object) :

    # ----------------------------------------------------------------
    #
    def __init__ (self, config, name, func_pre, func_core, func_post) :

        try:
            import numpy
        except ImportError, e:
            raise RuntimeError ("Benchmark depends on numpy which is not installed")
    
        self.pre    = func_pre
        self.core   = func_core
        self.post   = func_post
        self.name   = name.replace (' ', '_')
        self.lock   = rut.RLock ()
        self.events = dict()
    
        if  isinstance (config, basestring) :
            app_cfg = rutest.TestConfig (config)
        elif isinstance (config, dict) :
            app_cfg = config
        else :
            print 'warning: no valid benchmark config (neither filename nor dict) -- ignore'
            app_cfg = dict()

        if  'benchmarks' in app_cfg :
            bench_cfg = app_cfg['benchmarks']
        else :
            bench_cfg = dict ()

    
        # RADICAL_BENCHMARK_ environments will overwrite config settings
        if  'RADICAL_BENCHMARK_TAGS' in os.environ :
            bench_cfg['tags'] = os.environ['RADICAL_BENCHMARK_TAGS'].split (' ')
    
        if  'RADICAL_BENCHMARK_CONCURRENCY' in os.environ :
            bench_cfg['concurrency'] = os.environ['RADICAL_BENCHMARK_CONCURRENCY']
    
        if  'RADICAL_BENCHMARK_ITERATIONS' in os.environ :
            bench_cfg['iterations'] = os.environ['RADICAL_BENCHMARK_ITERATIONS']
    
        
        # check benchmark settings for completeness, set some defaults
        if  not 'concurrency' in bench_cfg : 
            bench_cfg['concurrency'] = 1
    
        if  not 'iterations'  in bench_cfg : 
            bench_cfg['iterations'] = 1
    
        if  not 'tags' in bench_cfg : 
            bench_cfg['tags'] = list()
    
        self.app_cfg   = app_cfg
        self.bench_cfg = bench_cfg
    

    # --------------------------------------------------------------------------
    #
    def _thread (self, tid) :
    
        try :
            self.pre (tid, self.app_cfg, self.bench_cfg)

            sys.stdout.write ('-')
            sys.stdout.flush ()
    
            self.events[tid]['event_1'].set  ()  # signal we are done        
            self.events[tid]['event_2'].wait ()  # wait 'til others are done 
    
            iterations = int(self.bench_cfg['iterations']) / int(self.bench_cfg['concurrency'])
    
            # poor-mans ceil()
            if (iterations * int(self.bench_cfg['concurrency'])) < int(self.bench_cfg['iterations']) :
                iterations += 1
    
            for i in range (0, iterations+1) :
                core_timer = self.core (tid, i, self.app_cfg, self.bench_cfg)
                self._tic (tid, core_timer)
    
    
            self.events[tid]['event_3'].set ()   # signal we are done        
            self.events[tid]['event_4'].wait ()  # wait 'til others are done 
    
    
            self.post (tid, self.app_cfg, self.bench_cfg)
            sys.stdout.write ('=')
            sys.stdout.flush ()
    
            self.events[tid]['event_5'].set ()   # signal we are done        

    
        except Exception as e :
    
            sys.stdout.write ("exception in benchmark thread: %s\n\n" % str(e))
            sys.stdout.write (traceback.format_exc())
            sys.stdout.write ("\n")
            sys.stdout.flush ()
    
            # Oops, we are screwed.  Tell main thread that we are done for, and
            # bye-bye...
            self.events[tid]['event_1'].set  ()
            self.events[tid]['event_3'].set  ()
            self.events[tid]['event_5'].set  ()

            raise

        # finish thread
        sys.exit (0)
    
    
    # --------------------------------------------------------------------------
    #
    def run (self) :
        """
        - create 'concurrency' number of threads
        - per thread call pre()
        - sync threads, start timer
        - per thread call core() 'iteration' number of times', tic()
        - stop timer
        - per thread, call post, close threads
        - eval once
        """
    
        threads     = []
        concurrency = int(self.bench_cfg['concurrency'])
    
        self._start ()
    
        for tid in range (0, concurrency) :
    
            self.events[tid] = {}
            self.events[tid]['event_1'] = rut.Event ()
            self.events[tid]['event_2'] = rut.Event ()
            self.events[tid]['event_3'] = rut.Event ()
            self.events[tid]['event_4'] = rut.Event ()
            self.events[tid]['event_5'] = rut.Event ()
            self.start [tid] = time.time ()
            self.times [tid] = list()
    
            t = rut.Thread (self._thread, tid)
            threads.append (t)
    
    
        for t in threads :
            t.start ()
    
        
        # wait for all threads to start up and initialize
        self.t_init = time.time ()
        rut.lout ("\n> " + "="*concurrency)
        rut.lout ("\n> ")
        for tid in range (0, concurrency) :
            self.events[tid]['event_1'].wait ()
    
        # start workload in all threads
        self.t_start = time.time ()
        for tid in range (0, concurrency) :
            self.events[tid]['event_2'].set ()
    
        # wait for all threads to finish core test
        for tid in range (0, concurrency) :
            self.events[tid]['event_3'].wait ()
        self.t_stop = time.time ()
    
        # start shut down
        rut.lout ("\n< " + "-"*concurrency)
        rut.lout ("\n< ")
        for tid in range (0, concurrency) :
            self.events[tid]['event_4'].set ()
    
        # wait for all threads to finish shut down
        for tid in range (0, concurrency) :
            self.events[tid]['event_5'].wait ()
    
    
    
    # ----------------------------------------------------------------
    #
    def _start (self) :
    
        self.start = dict()
        self.times = dict()
        self.idx   = 0
    
        rut.lout ("\n")
        rut.lout ("benchmark   : %s (%s)\n" % (self.name, self.bench_cfg['tags']))
        rut.lout ("concurrency : %s\n"      %  self.bench_cfg['concurrency'])
        rut.lout ("iterations  : %s\n"      %  self.bench_cfg['iterations'])
    
        sys.stdout.flush ()
    
    
    # ----------------------------------------------------------------
    #
    def _tic (self, tid='master_tid', core_timer=None) :
        # if a core_timer is given, we interpret that as time value instead of
        # using 'now-start'.  This assumes that 'core' has its own way of
        # measuring the time, possibly to correct for some systematic error...
    
        with self.lock :

            import numpy

            now   = time.time ()

            if  core_timer :
                timer = core_timer
            else :
                timer = now - self.start[tid]

            self.times[tid].append (timer)
            self.start[tid] = now

            if len(self.times[tid][1:]) :
                numpy_times = numpy.array (self.times[tid][1:])
                vmean = numpy_times.mean ()
            else :
                vmean = timer
    
            if   timer  <  0.75 * vmean : marker = '='
            if   timer  <  0.90 * vmean : marker = '~'
            elif timer  <  0.95 * vmean : marker = '_'
            elif timer  <  1.05 * vmean : marker = '.'
            elif timer  <  1.10 * vmean : marker = '-'
            elif timer  <  1.25 * vmean : marker = '+'
            else                        : marker = '*'
    
    
            if       not ( (self.idx)        ) : sys.stdout.write ('\n* ')
            else :
                if   not ( (self.idx) % 1000 ) : sys.stdout.write (" %7d\n\n# " % self.idx)
                elif not ( (self.idx) %  100 ) : sys.stdout.write (" %7d\n| "   % self.idx)
                elif not ( (self.idx) %   10 ) : sys.stdout.write (' ')
            if  True                                    : sys.stdout.write (marker)
            
            sys.stdout.flush ()
    
            self.idx += 1
    
    # ----------------------------------------------------------------
    #
    def eval (self, error=None) :
    
        import numpy

        times = list()
    
        for tid in self.times :
            times += self.times[tid][1:]

      # import pprint
      # pprint.pprint (times)
    
        if  len(times) < 1 :
            raise ValueError ("min 1 timing value required for benchmark evaluation (%d)" % len(times))
    
        concurrency = int(self.bench_cfg['concurrency'])
        tags        = ' '.join (self.bench_cfg['tags'])
    
        out = "\n"
        top = ""
        tab = ""
        num = ""
    
        out += "Results :\n"

        numpy_times = numpy.array (times)
    
        vtot  = self.t_stop  - self.t_start
        vini  = self.t_start - self.t_init
        vn    = len (times)
        vsum  = sum (times)
        vmin  = min (times)
        vmax  = max (times)

        if len (times) :
            vmean = numpy_times.mean ()
            vsdev = numpy_times.std  ()
        else :
            vmean = 0.0
            vsdev = 0.0
        vrate = vn / vtot
    
        bdat  = "benchmark.%s.dat" % (self.name)
    
        out += "  name    : %s\n"                               % (self.name)
        out += "  tags    : %s\n"                               % (tags)
        out += "  threads : %8d        iters   : %8d\n"        % (concurrency, vn)
        out += "  init    : %8.2fs       min     : %8.2fs\n"    % (vini,        vmin )
        out += "  total   : %8.2fs       max     : %8.2fs\n"    % (vtot,        vmax )
        out += "  rate    : %8.2f/s      mean    : %8.2fs\n"    % (vrate,       vmean)
        out += "                            sdev    : %8.2fs\n" % (             vsdev)
    
        num = "# %7s  %7s  %7s  %7s  %7s  %7s  %8s  %8s  %9s  %10s  %10s" \
            % (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
        top = "# %7s  %7s  %7s  %7s  %7s  %7s  %8s  %8s  %9s  %10s  %10s" \
            % ('n', 'threads', 'init', 'tot', 'min',  'max', 'mean', \
               'std-dev', 'rate', 'name', 'tags')
    
        tab = "  "      \
              "%7d  "   \
              "%7d  "   \
              "%7.2f  " \
              "%7.2f  " \
              "%7.2f  " \
              "%7.2f  " \
              "%8.3f  " \
              "%8.3f  " \
              "%9.3f  " \
              "%10s  "  \
              "%10s  "  \
            % (vn, 
               concurrency, 
               vini,
               vtot,   
               vmin,  
               vmax, 
               vmean, 
               vsdev, 
               vrate, 
               self.name, 
               tags) 
    
        rut.lout ("\n%s" % out)
    
        create_top = True
        try :
            statinfo = os.stat (bdat)
            if  statinfo.st_size > 0 :
                create_top = False
        except :
            pass
    
        f = open (bdat, "a+")
    
        if  create_top :
            f.write ("%s\n" % num)
            f.write ("%s\n" % top)
        f.write ("%s\n" % tab)
    
    
# --------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys
import time
import socket
import threading       as mt
import multiprocessing as mp
import setproctitle    as spt

from .logger  import get_logger
from .threads import Thread as ru_Thread
from .        import poll   as ru_poll


# ------------------------------------------------------------------------------
#
_ALIVE_MSG     = 'alive'  # message to use as alive signal
_START_TIMEOUT = 20.0     # time to wait for process startup signal.
                          # startup signal: 'alive' message on the socketpair;
                          # is sent in both directions to ensure correct setup
_WATCH_TIMEOUT = 0.2      # time between thread and process health polls.
                          # health poll: check for recv, error and abort
                          # on the socketpair; is done in a watcher thread.
_STOP_TIMEOUT  = 2.0     # time between temination signal and killing child
_BUFSIZE       = 1024     # default buffer size for socket recvs


# ------------------------------------------------------------------------------
#
class Process(mp.Process):
    '''
    This `Process` class is a thin wrapper around `mp.Process` which
    specifically manages the process lifetime in a more cautious and copperative
    way than the base class: *no* attempt on signal handling is made, we expect
    to exclusively communicate between parent and child via a socket.
    A separate thread in both processes will watch that socket: if the socket
    disappears, we interpret that as the other process being terminated, and
    begin process termination, too.

    NOTE: At this point we do not implement the full `mp.Process` constructor.

    The class also implements a number of initialization and finalization
    methods which can be overloaded by any deriving class.  While this can at
    first be a confusing richness of methods to overload, it significantly
    simplifies the implementation of non-trivial child processes.  By default,
    none of the initialized and finalizers needs to be overloaded.

    An important semantic difference are the `start()` and `stop()` methods:
    both accept an optional `timeout` parameter, and both guarantee that the
    child process successfully started and completed upon return, respectively.
    If that does not happen within the given timeout, an exception is raised.
    Not that the child startup is *only* considered complete once all of its
    initialization methods have been completed -- the start timeout value must
    thus be chosen very carefully.  Note further that the *parent*
    initialization methods are also part of the `start()` invocation, and must
    also be considered when choosing the timeout value.  Parent initializers
    will only be executed once the child is known to be alive, and the parent
    can thus expect the child bootstrapping to be completed (avoiding the need
    for additional handshakes etc).  Any error in child or parent initialization
    will result in an exception, and the child will be terminated.

    Along the same lines, the parent and child finalizers are executed in the
    `stop()` method, prompting similar considerations for the `timeout` value.

    Any class which derives from this Process class *must* overload the 'work_cb()`
    method.  That method is repeatedly called by the child process' main loop,
    until:
      - an exception occurs (causing the child to fail with an error)
      - `False` is returned by `work_cb()` (causing the child to finish w/o error)
    '''

    # TODO: We should switch to fork/*exec*, if possible.


    # --------------------------------------------------------------------------
    #
    def __init__(self, name, log=None):

        # At this point, we only initialize members which we need before start
        self._ru_name      = name  # use for setproctitle
        self._ru_childname = self._ru_name + '.child'

        # make sure we have a valid logger
        self._ru_set_logger(log)

        # ... or are *shared* between parent and child process.
        self._ru_ppid = os.getpid()    # get parent process ID

        # most importantly, we create a socketpair.  Both parent and child will
        # watch one end of that socket, which thus acts as a lifeline between
        # the processes, to detect abnormal termination in the process tree.
        # The socket is also used to send messages back and forth.
        self._ru_sp = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)

        # note that some members are only really initialized when calling the
        # `start()`/`run()` methods, and the various initializers.
        self._ru_spawned     = None           # set in start(), run()
        self._ru_is_parent   = None           # set in start()
        self._ru_is_child    = None           # set in run()
        self._ru_endpoint    = None           # socket endpoint for sent/recv
        self._ru_term        = None           # set to terminate watcher
        self._ru_initialized = False          # set to signal bootstrap success
        self._ru_terminating = False          # set to signal active termination
        self._ru_watcher     = None           # watcher thread
        self._ru_things_lock = mt.Lock()      # lock for the above
        self._ru_things      = dict()         # registry of threads created in
                                              # (parent or child) process

        # FIXME: assert that start() was called for some / most methods

        # when cprofile is requested but not available,
        # we complain, but continue unprofiled
        self._ru_cprofile = False
        if self._ru_name in os.environ.get('RADICAL_CPROFILE', '').split(':'):
            try:
                self._ru_log.error('enable cprofile for %s', self._ru_name)
                import cprofile
                self._ru_cprofile = True
            except:
                self._ru_log.error('cannot import cprofile - disable')

        # base class initialization
        super(Process, self).__init__(name=self._ru_name)

        # we don't want processes to linger around, waiting for children to
        # terminate, and thus create sub-processes as daemons.
        #
      # NOTE: this requires the `patchy.mc_patchface()` fix in `start()`.
      #
      # self.daemon = True


    # --------------------------------------------------------------------------
    #
    @property
    def ru_name(self):
        return self._ru_name


    # --------------------------------------------------------------------------
    #
    @property
    def ru_childname(self):
        return self._ru_childname


    # --------------------------------------------------------------------------
    #
    def _ru_set_logger(self, log=None):

        self._ru_log = log

        if not self._ru_log:
            # if no logger is passed down, log to null (FIXME)
            self._ru_log = get_logger('radical.util.process', target='2')
            self._ru_log.debug('name: %s' % self._ru_name)


    # --------------------------------------------------------------------------
    #
    def _ru_msg_send(self, msg):
        '''
        send new message to self._ru_endpoint.  We make sure that the
        message is not larger than the _BUFSIZE define in the RU source code.
        '''

        if not self._ru_spawned:
            # no child, no communication channel
            info = "=== = %s\n=== = can't communicate w/o child %s -> %s [%s]" \
                    % (msg, os.getpid(), self.pid, self._ru_name)
            self._ru_log.debug(info)
            print info
            raise RuntimeError("can't communicate w/o child (%s -> %s [%s]",
                    os.getpid(), self.pid, self._ru_name)

        # NOTE:  this method should only be called by the watcher thread, which
        #        owns the endpoint.
        # FIXME: BUFSIZE should not be hardcoded
        # FIXME: check for socket health

        if len(msg) > _BUFSIZE:
            raise ValueError('message is larger than %s: %s' % (_BUFSIZE, msg))

        self._ru_log.info('send message: [%s] %s', self._ru_name, msg)
        try:
            self._ru_endpoint.send(msg)
        except Exception as e:
            # this should only happen once the EP is done for - terminate
            self._ru_log.warn('send failed (%s) - terminate', e)
            if self._ru_term is not None:
                self._ru_term.set()



    # --------------------------------------------------------------------------
    #
    def _ru_msg_recv(self, size=_BUFSIZE, timeout=None):
        '''
        receive a message from self._ru_endpoint.  We only check for messages
        of *up to* `size`.

        This call is non-blocking: if no message is available, return an empty
        string.
        '''

        if not self._ru_spawned:
            # no child, no communication channel
            raise RuntimeError("can't communicate w/o child")

        # NOTE:  this method should only be called by the watcher thread, which
        #        owns the endpoint (no lock used!)
        # FIXME: BUFSIZE should not be hardcoded
        # FIXME: check for socket health

        try:
            if timeout:
                self._ru_endpoint.settimeout(timeout)
            msg = self._ru_endpoint.recv(size)
            self._ru_log.info('recv message: %s', msg)
            return msg

        except socket.timeout:
            self._ru_log.warn('recv timed out')
            return ''
        except Exception as e:
            # this should only happen once the EP is done for - terminate
            self._ru_log.warn('recv failed (%s) - terminate', e)
            self._ru_term.set()


    # --------------------------------------------------------------------------
    #
    def _ru_watch(self):
        '''
        When `start()` is called, the parent process will create a socket pair.
        after fork, one end of that pair will be watched by the parent and
        client, respectively, in separate watcher threads.  If any error
        condition or hangup is detected on the socket, it is assumed that the
        process on the other end died, and termination is initiated.

        Since the watch happens in a subthread, any termination requires the
        attention and cooperation of the main thread.  No attempt is made on
        interrupting the main thread, we only set self._ru_term which needs to
        be checked by the main threads in certain intervals.
        '''

        # we watch sockets and threads as long as we live, ie. until the main
        # thread sets `self._ru_term`.
        try:

            self._ru_poller = ru_poll.poll(logger=self._ru_log)
            self._ru_poller.register(self._ru_endpoint,
                    ru_poll.POLLIN  | ru_poll.POLLERR | ru_poll.POLLHUP)
                 #  ru_poll.POLLPRI | ru_poll.POLLNVAL)

            last = 0.0  # we never watched anything until now
            while not self._ru_term.is_set() :

                # only do any watching if time is up
                now = time.time()
                if now - last < _WATCH_TIMEOUT:
                    time.sleep(0.1)  # FIXME: configurable, load tradeoff
                    continue

                if  not self._ru_watch_socket() or \
                    not self._ru_watch_things()    :
                    return

                last = now

                # FIXME: also *send* any pending messages to the child.
              # # check if any messages need to be sent.
              # while True:
              #     try:
              #         msg = self._ru_msg_out.get_nowait()
              #         self._ru_msg_send(msg)
              #
              #     except Queue.Empty:
              #         # nothing more to send
              #         break


        except Exception as e:
            # mayday... mayday...
            self._ru_log.exception('watcher failed')

        finally:
            # no matter why we fell out of the loop: let the other end of the
            # socket know by closing the socket endpoint.

            # Fix radical-cybertools/radical.utils issue #120
            # We need to call SHUTDOWN before we close the socket...
            # From: https://docs.python.org/2/howto/sockets.html#disconnecting
            self._ru_log.info('watcher closes')
            self._ru_endpoint.shutdown(socket.SHUT_RDWR)
            self._ru_endpoint.close()
            self._ru_poller.close()

            # `self.stop()` will be called from the main thread upon checking
            # `self._ru_term` via `self.is_alive()`.
            # FIXME: check
            self._ru_term.set()


    # --------------------------------------------------------------------------
    #
    def _ru_watch_socket(self):

        # check health of parent/child relationship
        events = self._ru_poller.poll(0.1)   # block just a little
        for _,event in events:

            # for alive checks, we poll socket state for
            #   * data:   some message from the other end, logged
            #   * error:  child failed   - terminate
            #   * hangup: child finished - terminate

            # check for messages
            if event & ru_poll.POLLIN:

                # we get a message!
                #
                # FIXME: BUFSIZE should not be hardcoded
                # FIXME: we do nothing with the message yet, should be
                #        stored in a message queue.

                msg = self._ru_msg_recv(_BUFSIZE)
                self._ru_log.info('message received: %s' % msg)
                if msg == '':
                    self._ru_log.warn('received empty string, parent closed ep')
                    return False

                if msg.strip() == 'STOP':
                    self._ru_log.info('STOP received: %s' % msg)
                    return False

            # check for error conditions
            if  event & ru_poll.POLLHUP or  \
                event & ru_poll.POLLERR     :

                # something happened on the other end, we are about to die
                # out of solidarity (or panic?).
                self._ru_log.warn('endpoint disappeard')
                return False

          # if event & ru_poll.POLLPRI:
          #     self._ru_log.info('POLLPRI : %s', self._ru_endpoint.fileno())
          # if event & ru_poll.POLLNVAL:
          #     self._ru_log.info('POLLNVAL: %s', self._ru_endpoint.fileno())

        time.sleep(0.1)

      # self._ru_log.debug('endpoint watch ok')
        return True


    # --------------------------------------------------------------------------
    #
    def register_watchable(self, thing):
        '''
        Add an object to watch.  If the object is at any point found to be not
        alive (`thing.is_alive()` returns `False`), an exception is raised which
        will cause the watcher thread to terminate, and will thus also terminate
        this `ru.Process` instance.  All registered things will be terminated on
        `stop()`.

        The given `thing` is expected to have these four methods/properties

          - `thing.is_alive()`  -- perform health check
          - `thing.name`        -- a unique name identifying the thing
          - `thing.stop()`      -- method to terminate thing on regular shutdown
          - `thing.join()`      -- method to collect   thing on regular shutdown

        `thing.stop()` and `thing.join()` are expected to accept a timeout
        parameter, and are expected to raise exceptions if the operation fails.

        In other words, `things` are really expected to be radical.utils threads
        and processes, but we don't enforce this in terms of type checks.
        '''

        assert(thing.is_alive)
        assert(thing.name)
        assert(thing.stop)
        assert(thing.join)

        with self._ru_things_lock:
            if thing.name in self._ru_things:
                raise ValueError('already watching %s' % thing.name)
            self._ru_things[thing.name] = thing


    # --------------------------------------------------------------------------
    #
    def unregister_watchable(self, name):

        with self._ru_things_lock:
            if name not in self._ru_things:
                raise ValueError('%s is not watched' % name)

            del(self._ru_things[name])


    # --------------------------------------------------------------------------
    #
    def _ru_watch_things(self):

        with self._ru_things_lock:
            for tname,thing in self._ru_things.iteritems():
                if not thing.is_alive():
                    self._ru_log.warn('%s died')
                    return False

        return True


    # --------------------------------------------------------------------------
    #
    def start(self, spawn=True, timeout=None):
        '''
        Overload the `mp.Process.start()` method, and block (with timeout) until
        the child signals to be alive via a message over our socket pair.  Also
        run all initializers.

        If spawn is set to `False`, then no child process is actually created,
        but the parent initializers will be executed nonetheless.

        '''

        self._ru_log.debug('start process')

        if timeout is None:
            timeout = _START_TIMEOUT

        # Daemon processes can't fork child processes in Python, because...
        # Well, they just can't.  We want to use daemons though to avoid hanging
        # processes if, for some reason, communication of termination conditions
        # fails.
        #
        # Patchy McPatchface to the rescue (no, I am not kidding): we remove
        # that useless assert (of all things!) on the fly.
        #
        # NOTE: while this works, we seem to have the socketpair-based detection
        #       stable enough to not need the monkeypatch.
        #
      # _daemon_fork_patch = '''\
      #     *** process_orig.py  Sun Nov 20 20:02:23 2016
      #     --- process_fixed.py Sun Nov 20 20:03:33 2016
      #     ***************
      #     *** 5,12 ****
      #           assert self._popen is None, 'cannot start a process twice'
      #           assert self._parent_pid == os.getpid(), \\
      #                  'can only start a process object created by current process'
      #     -     assert not _current_process._daemonic, \\
      #     -            'daemonic processes are not allowed to have children'
      #           _cleanup()
      #           if self._Popen is not None:
      #               Popen = self._Popen
      #     --- 5,10 ----
      #     '''
      #
      # import patchy
      # patchy.mc_patchface(mp.Process.start, _daemon_fork_patch)

        if spawn:
            # start `self.run()` in the child process, and wait for it's
            # initalization to finish, which will send the 'alive' message.
            super(Process, self).start()
            self._ru_spawned = True


        # this is the parent now - set role flags.
        self._ru_is_parent = True
        self._ru_is_child  = False

        # select different ends of the socketpair for further communication.
        self._ru_endpoint  = self._ru_sp[0]
        self._ru_sp[1].close()

        # from now on we need to invoke `self.stop()` for clean termination.
        # Having said that: the daemonic watcher thread and the socket lifeline
        # to the child should ensure that both will terminate in all cases, but
        # possibly somewhat delayed and abruptly.
        #
        # Either way: use a try/except to ensure `stop()` being called.
        try:

            if self._ru_spawned:
                # we expect an alive message message from the child, within
                # timeout
                #
                # NOTE: If the child does not bootstrap fast enough, the timeout
                #       will kick in, and the child will be considered dead,
                #       failed and/or hung, and will be terminated!  Timeout can
                #       be set as parameter to the `start()` method.
                msg = self._ru_msg_recv(size=len(_ALIVE_MSG), timeout=timeout)

                if not msg:
                    raise RuntimeError('child %s failed to come up [%ss]' %
                                       (self._ru_childname, msg))

                elif msg != _ALIVE_MSG:
                    # attempt to read remainder of message and barf
                    msg += self._ru_msg_recv()
                    raise RuntimeError('%s got unexpected message (%s) [%s]' % 
                                       (self._ru_name, msg, timeout))


            # When we got the alive messages, only then will we call the parent
            # initializers.  This way those initializers can make some
            # assumptions about successful child startup.
            self._ru_initialize()

        except Exception as e:
            self._ru_log.exception('initialization failed (%s)' % e)
            self.stop()
            raise

        # if we got this far, then all is well, we are done.
        if self._ru_spawned:
            self._ru_log.debug('process class started child')
        else:
            self._ru_log.debug('process class started (no child)')

        # child is alive and initialized, parent is initialized, watcher thread
        # is started - Wohoo!


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        This method MUST NOT be overloaded!

        This is the workload of the child process.  It will first call the child
        initializers and then repeatedly call `self.work_cb()`, until being
        terminated.  When terminated, it will call the child finalizers and
        exit.  Note that the child will also terminate once `work_cb()` returns
        `False`.

        The implementation of `work_cb()` needs to make sure that this process is
        not spinning idly -- if there is nothing to do in `work_cb()` at any point
        in time, the routine should at least sleep for a fraction of a second or
        something.

        Child finalizers are only guaranteed to get called on `self.stop()` --
        a hard kill via `self.terminate()` may or may not be able to trigger to
        run the child finalizers.

        The child process will automatically terminate (incl. finalizer calls)
        when the parent process dies. It is thus not possible to create daemon
        or orphaned processes -- which is an explicit purpose of this
        implementation.
        '''

        # if no profiling is wanted, we just run the workload and exit
        if not self._ru_cprofile:
            self._run()

        # otherwise we run under the profiler, obviously
        else:
            import cprofile
            cprofiler = cprofile.Profile()
            cprofiler.runcall(self._run)
            cprofiler.dump_stats('%s.cprof' % (self._ru_name))


    # --------------------------------------------------------------------------
    #
    def _run(self):

        # FIXME: ensure that this is not overloaded

        # this is the child now - set role flags
        self._ru_is_parent = False
        self._ru_is_child  = True
        self._ru_spawned   = True

        # select different ends of the socketpair for further communication.
        self._ru_endpoint  = self._ru_sp[1]
        self._ru_sp[0].close()

        # set child name based on name given in c'tor, and use as proctitle
        self._ru_name = self._ru_childname
        spt.setproctitle(self._ru_name)

        try:
            # we consider the invocation of the child initializers to be part of
            # the bootstrap process, which includes starting the watcher thread
            # to watch the parent's health (via the socket health).
            try:
                self._ru_initialize()

            except BaseException as e:
                self._ru_log.exception('abort: %s', repr(e))
                self._ru_msg_send('error: %s' % repr(e))
              # sys.stderr.write('initialization error in %s: %s\n' % (self._ru_name, repr(e)))
              # sys.stderr.flush()
                if self._ru_term is not None:
                    self._ru_term.set()

            # initialization done - we only now send the alive signal, so the
            # parent can make some assumptions about the child's state
            if not self._ru_term.is_set():
                self._ru_log.info('send alive')
                self._ru_msg_send(_ALIVE_MSG)

            # enter the main loop and repeatedly call 'work_cb()'.
            #
            # If `work_cb()` ever returns `False`, we break out of the loop to call the
            # finalizers and terminate.
            #
            # In each iteration, we also check if the socket is still open -- if it
            # is closed, we assume the parent to be dead and terminate (break the
            # loop).  We consider the socket closed if `self._ru_term` was set
            # by the watcher thread.
            while not self._ru_term.is_set() and \
                      self._parent_is_alive()    :

                # des Pudel's Kern
                if not self.work_cb():
                    self._ru_msg_send('work finished')
                    break

        except BaseException as e:

            # This is a very global except, also catches
            # sys.exit(), keyboard interrupts, etc.
            # Ignore pylint and PEP-8, we want it this way!
            self._ru_log.exception('abort: %s', repr(e))
            try:
                self._ru_msg_send('abort: %s' % repr(e))
            except Exception as e:
                self._ru_log.exception('abort info not sent: %s', repr(e))

        try:
            # note that we always try to call the finalizers, even if an
            # exception got raised during initialization or in the work loop
            # initializers failed for some reason or the other...
            self._ru_finalize()

        except BaseException as e:
            self._ru_log.error('finalization error: %s' % repr(e))
            try:
                self._ru_msg_send('finalize: %s' % repr(e))
            except Exception as e2:
                self._ru_log.exception('finalize error not sent: %s', repr(e2))

        try:
            self._ru_msg_send('terminating')
        except Exception as e:
            self._ru_log.warn('term msg error not sent: %s', repr(e))

        # tear down child watcher
        if self._ru_watcher:
            self._ru_term.set()
            self._ru_watcher.join(_STOP_TIMEOUT)

        # stop all things we watch
        with self._ru_things_lock:
            for tname,thing in self._ru_things.iteritems():
                try:
                    thing.stop(timeout=_STOP_TIMEOUT)
                except Exception as e:
                    self._ru_log.exception('Could not stop %s [%s]', tname, e)

        # All watchables should have stopped.  For some of them,
        # `stop()` will have implied `join()` already - but an additional
        # `join()` will cause little overhead, so we don't bother
        # distinguishing.

        with self._ru_things_lock:
            for tname,thing in self._ru_things.iteritems():
                try:
                    thing.join(timeout=_STOP_TIMEOUT)
                except Exception as e:
                    self._ru_log.exception('3 could not join %s [%s]', tname, e)

        # all is done and said - begone!
        sys.exit(0)


    # --------------------------------------------------------------------------
    #
    def stop(self, timeout=None):
        '''
        `stop()` is symmetric to `start()`, in that it can only be called by the
        parent process.  It MUST be called from the main thread.  Both
        conditions are asserted.  If a subthread or the child needs to trigger
        the termination of the parent, it should simply terminate/exit its own
        scope, and let the parent detect that via its watcher thread.  That will
        eventually cause `is_alive()` to return `False`, signalling the
        application that `stop()` should be called.

        This method sets the thread termination signal, and call `stop()` on all
        watchables.  It then calls `join()` on all watchables, with the given
        timeout, and then also joins the child process with the given timeout.

        If any of the join calls fails or times out, an exception will be
        raised.  All join's will be attempted though, to collect as many threads
        and processes as possible.


        NOTE: `stop()` implies `join()`!  Use `terminate()` if that is not
              wanted.  Terminate will not stop watchables though.

        NOTE: The given timeout is, as described above, applied multiple times,
        once for the child process, and once for each watchable.
        '''

        # FIXME: we can't really use destructors to make sure stop() is called,
        #        but we should consider using `__enter__`/`__leave__` scopes to
        #        ensure clean termination.

        # FIXME: This method should reduce to
        #           self.terminate(timeout)
        #           self.join(timeout)
        #        ie., we should move some parts to `terminate()`.

        if not timeout:
            timeout = _STOP_TIMEOUT

        # make sure we don't recurse
        if self._ru_terminating:
            return
        self._ru_terminating = True

      # if not is_main_thread():
      #     self._ru_log.info('reroute stop to main thread (%s)' % self._ru_name)
      #     sys.exit()

        self._ru_log.info('parent stops child  %s -> %s [%s]', os.getpid(),
                self.pid, self._ru_name)
        if self._ru_spawned:
            self._ru_msg_send('STOP')

        # keep a list of error messages for an eventual exception message
        errors = list()

        # call parent finalizers
        self._ru_finalize()

        # tear down watcher - we wait for it to shut down, to avoid races
        if self._ru_watcher:
            self._ru_term.set()
            self._ru_watcher.stop(timeout)

        # stop all things we watch
        with self._ru_things_lock:
            for tname,thing in self._ru_things.iteritems():
                try:
                    thing.stop(timeout=timeout)
                except Exception as e:
                    errors.append('could not stop %s [%s]' % (tname, e))

        # process termination is only relevant for the parent, and only if
        # a child was actually spawned
        if self._ru_is_parent and self._ru_spawned:

            # stopping the watcher will close the socket, and the child should
            # begin termination immediately.  Well, whenever it realizes the
            # socket is gone, really.  We wait for that termination to complete.
            super(Process, self).join(timeout)

            # make sure child is gone
            if super(Process, self).is_alive():
                self._ru_log.warn('failed to stop child - terminate: %s -> %s [%s]', os.getpid(), self.pid, self._ru_name)
                self.terminate()  # hard kill
                super(Process, self).join(timeout)

          # # check again
          # if super(Process, self).is_alive():
          #     # we threat join errors as non-fatal here - at this point, there
          #     # is not much we can do other than calling `terminate()#join()`
          #     # -- which is exactly what we just did.
          #     self._ru_log.warn('could not join child process %s', self.pid)

        # meanwhile, all watchables should have stopped, too.  For some of them,
        # `stop()` will have implied `join()` already - but an additional
        # `join()` will cause little overhead, so we don't bother
        # distinguishing.
        with self._ru_things_lock:
            for tname,thing in self._ru_things.iteritems():
                try:
                    thing.join(timeout=timeout)
                except Exception as e:
                    self._ru_log.exception('could not join %s [%s]', tname, e)
                    errors.append('could not join %s [%s]' % (tname, e))

        # don't exit - but signal if the child or any watchables survived
        if errors:
            for error in errors:
                self._ru_log.error(error)
            raise RuntimeError(errors)


    # --------------------------------------------------------------------------
    #
    def join(self, timeout=None):

      # raise RuntimeError('call stop instead!')
      #
      # we can't really raise the exception above, as the mp module calls this
      # join via `at_exit`.  Which kind of explains hangs on unterminated
      # children...
      #
      # FIXME: not that `join()` w/o `stop()` will not call the parent finalizers.
      #        We should call those in both cases, but only once.
      # FIXME: `join()` should probably block by default

        if not timeout:
            timeout = _STOP_TIMEOUT

        assert(self._ru_is_parent)

        if self._ru_spawned:
            super(Process, self).join(timeout=timeout)


    # --------------------------------------------------------------------------
    #
    def _ru_initialize(self):
        '''
        Perform basic settings, then call common and parent/child initializers.
        '''

        try:
            # private initialization
            self._ru_term = mt.Event()

            # call parent and child initializers, respectively
            if self._ru_is_parent:
                self.ru_initialize_common()
                self.ru_initialize_parent()

                self._ru_initialize_common()
                self._ru_initialize_parent()

            elif self._ru_is_child:
                self.ru_initialize_common()
                self.ru_initialize_child()

                self._ru_initialize_common()
                self._ru_initialize_child()

            self._ru_initialized = True

        except Exception as e:
            self._ru_log.exception('initialization error')
            raise


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_common(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_parent(self):

        if not self._ru_spawned:
            # no child, so we won't need a watcher either
            return

        # Start a separate thread which watches our end of the socket.  If that
        # thread detects any failure on that socket, it will set
        # `self._ru_term`, to signal its demise and prompt an exception from
        # the main thread.
        #
        # NOTE: For several reasons, the watcher thread has no valid/stable
        #       means of directly signaling the main thread of any error
        #       conditions, it is thus necessary to manually check the child
        #       state from time to time, via `self.is_alive()`.
        #
        # NOTE: https://bugs.python.org/issue1856  (01/2008)
        #       `sys.exit()` can segfault Python if daemon threads are active.
        #       https://bugs.python.org/issue21963 (07/2014)
        #       This will not be fixed in python 2.x.
        #
        #       We make the Watcher thread a daemon anyway:
        #
        #       - when `sys.exit()` is called in a child process, we don't care
        #         about the process anymore anyway, and all terminfo data are
        #         sent to the parent anyway.
        #       - when `sys.exit()` is called in the parent on unclean shutdown,
        #         the same holds.
        #       - when `sys.exit()` is called in the parent on clean shutdown,
        #         then the watcher threads should already be terminated when the
        #         `sys.exit()` invocation happens
        #
        # FIXME: check the conditions above
        #
        # FIXME: move to _ru_initialize_common
        #
        self._ru_watcher = ru_Thread(name='%s.watch' % self._ru_name,
                                     target=self._ru_watch,
                                     log=self._ru_log)
        self._ru_watcher.start()

        self._ru_log.info('child is alive')


    # --------------------------------------------------------------------------
    #
    def _ru_initialize_child(self):

        # TODO: should we also get an alive from parent?
        #
        # FIXME: move to _ru_initialize_common
        #

        self._ru_log.info('child (me) initializing')

        # start the watcher thread
        self._ru_watcher = ru_Thread(name='%s.watch' % self._ru_name,
                                     target=self._ru_watch,
                                     log=self._ru_log)
        self._ru_watcher.start()

        self._ru_log.info('child (me) is alive')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_common(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, for both the parent and the child process (individually).  If
        this fails on either side, the process startup is considered failed.
        '''

        self._ru_log.debug('ru_initialize_common (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_parent(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, in the parent process.  If this fails, the process startup is
        considered failed.
        '''

        self._ru_log.debug('ru_initialize_parent (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_initialize_child(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `start()`, in the child process.  If this fails, the process startup is
        considered failed.
        '''

        self._ru_log.debug('ru_initialize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def _ru_finalize(self):
        '''
        Call common and parent/child initializers.

        Note that finalizers are called in inverse order of initializers.
        '''

        try:
            # call parent and child finalizers, respectively
            if self._ru_is_parent:
                self.ru_finalize_parent()
                self.ru_finalize_common()

                self._ru_finalize_parent()
                self._ru_finalize_common()

            elif self._ru_is_child:
                self.ru_finalize_child()
                self.ru_finalize_common()

                self._ru_finalize_child()
                self._ru_finalize_common()

        except Exception as e:
            self._ru_log.exception('finalization error')
            raise RuntimeError('finalize: %s' % repr(e))


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_common(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_parent(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _ru_finalize_child(self):

        pass


    # --------------------------------------------------------------------------
    #
    def ru_finalize_common(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or process child termination, in the parent process, in both
        the parent and the child process (individually).
        '''

        self._ru_log.debug('ru_finalize_common (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_finalize_parent(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or process child termination, in the parent process.
        '''

        self._ru_log.debug('ru_finalize_parent (NOOP)')


    # --------------------------------------------------------------------------
    #
    def ru_finalize_child(self):
        '''
        This method can be overloaded, and will then be executed *once* during
        `stop()` or process child termination, in the child process.
        '''

        self._ru_log.debug('ru_finalize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def work_cb(self):
        '''
        This method MUST be overloaded.  It represents the workload of the
        process, and will be called over and over again.

        This has several implications:

          * `work_cb()` needs to enforce any call rate limits on its own!
          * in order to terminate the child, `work_cb()` needs to either raise an
            exception, or call `sys.exit()` (which actually also raises an
            exception).

        Before the first invocation, `self.ru_initialize_child()` will be called.
        After the last invocation, `self.ru_finalize_child()` will be called, if
        possible.  The latter will not always be possible if the child is
        terminated by a signal, such as when the parent process calls
        `child.terminate()` -- `child.stop()` should be used instead.

        The overloaded method MUST return `True` or `False` -- the child will
        continue to work upon `True`, and otherwise (on `False`) begin
        termination.
        '''

        raise NotImplementedError('ru.Process.work_cb() MUST be overloaded')


    # --------------------------------------------------------------------------
    #
    def is_alive(self, strict=True):
        '''
        Check if the child process is still alive, and also ensure that
        termination is not yet initiated.  If `strict` is set (default), then
        only the process state is checked.
        '''

        if not self._ru_spawned:
            # its not an error if the child was never spawned.
            alive = True

        elif self._ru_is_child:
            # if this  *is* the child, then it's alive
            alive = True

        else:
            # this is the parent, and it spawned: check for real
            alive = super(Process, self).is_alive()
            if not alive:
                self._ru_log.warn('super: alive check failed [%s]', alive)

        if self._ru_term is None:
            # child is not yet started
            self._ru_log.warn('startup: alive check failed [%s]', alive)
            return False

        termed = self._ru_term.is_set()


        if strict: return alive
        else     : return alive and not termed


    # --------------------------------------------------------------------------
    #
    def is_valid(self, term=True):
        '''
        This method provides a self-check, and will call `stop()` if that check
        fails and `term` is set.  If `term` is not set, the return value of
        `is_alive()` is passed through.

        The check will basically assert that `is_alive()` is `True`.   The
        purpose is to call this check frequently, e.g. at the begin of each
        method invocation and during longer loops, as to catch failing
        sub-processes or threads and to then terminate.

        This method will always return True if `start()` was not called, yet.
        '''

        if self._ru_terminating or not self._ru_initialized:
            self._ru_log.debug('alive check in term')
            return True

        alive = self.is_alive(strict=False)

        if not alive and term:
            self._ru_log.warn('alive check: proc invalid - stop [%s - %s]', alive, term)
            self.stop()
        else:
            return alive


    # --------------------------------------------------------------------------
    #
    def _parent_is_alive(self):
        '''
        This private method checks if the parent process is still alive.  This
        obviously only makes sense when being called in the child process.

        Note that there exists a (however unlikely) race: PIDs are reused, and
        the process could be replaced by another process with the same PID
        inbetween tests.  We thus also except *all* exception, including
        permission errors, to capture at least some of those races.
        '''

        # This method is an additional fail-safety check to the socketpair
        # watching performed by the watcher thread -- either one should
        # actually suffice.

        assert(self._ru_is_child)

        try:
            os.kill(self._ru_ppid, 0)
            return True

        except:
            return False


# ------------------------------------------------------------------------------

